import uuid
from abc import ABCMeta, abstractmethod
import datetime
import os
from intelliw.utils.storage_service import StorageService
from intelliw.datasets.datasource_base import AbstractDataSourceWriter, DataSourceWriterException, DataSourceType as DST
from intelliw.datasets.utils.iwfactory_dr_api import IwFactoryDrAPI
from intelliw.utils import iuap_request
from intelliw.utils.logger import _get_framework_logger
from intelliw.config import config
import pandas as pd
import numpy as np
import tempfile
import json
import traceback

logger = _get_framework_logger()


class OutputCheckException(Exception):
    def __init__(self, msg) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return '数据输出校验失败:\n错误原因: {}'.format(self.msg) + '''
发生错误的可能：
    1 返回字段是否与数据库定一致
    2 返回字段的列名是否与每行一一对应
    3 如果使用数据输出功能，返回数据格式应为：
        {
            "meta":[
                {"code":"col1"},
                {"code":"col2"},
                {"code":"col3"}
            ],
            "result":[
                ["col1_value1", "col2_value1", "col3_value1"],
                ["col1_value2", "col2_value2", "col3_value2"],
                ["col1_value3", "col2_value3", "col3_value3"],
            ]
        }'''

    def ignore_stack(self):
        return True


class OutputConfigException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self) -> str:
        return '数据输出配置文件参数错误，请检查参数:\n{}'.format(self.msg)

    def ignore_stack(self):
        return True


class OutputRequestException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self) -> str:
        return '数据输出接口请求出错:\n{}'.format(self.msg)

    def ignore_stack(self):
        return True


class AbstractEngineWriter(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.table_columns = None
        self.write_successed = {"status": 1}
        self.write_failed = {"status": 0}

    @abstractmethod
    def write(self, data, **kwargs):
        pass

    @abstractmethod
    def check_metadata(self, columns, **kwargs):
        '''
        校验建表字段是否合规
        {
            "code": "字段名",
            "type": "字段类型",
            "label": "字段描述"
        }
        :param columns: 字段列表
        :return: 返回字段合规结果
        '''
        pass

    @abstractmethod
    def create_table(self, table_name, columns, **kwargs):
        pass

class UserWriter(AbstractEngineWriter):
    def write(self, data, **kwargs):
        pass

    def check_metadata(self, columns, **kwargs):
        pass
    
    def create_table(self, table_name, columns, **kwargs):
        pass


class SqlWriter(AbstractEngineWriter):
    def __init__(self) -> None:
        super().__init__()
        self.uri = ""
        self.table_name = ""
        self.batch_size = 10000
        self.driver = ""
        self.engine = self.__get_engine()

    def __get_engine(self):
        from intelliw.utils.database import connection
        connection.get_engine(self.driver, uri=self.uri)

    def write(self, data, **kwargs):
        frame = pd.DataFrame(data["result"], columns=data["mate"])
        start_index = 0
        end_index = self.batch_size if self.batch_size < len(
            frame) else len(frame)

        frame = frame.where(pd.notnull(frame), None)
        if_exists_param = 'replace'

        while start_index != end_index:
            logger.info("Writing rows %s through %s" %
                        (start_index, end_index))
            frame.iloc[start_index:end_index, :].to_sql(
                con=self.engine, name=self.table_name, if_exists=if_exists_param)
            if_exists_param = 'append'

            start_index = min(start_index + self.batch_size, len(frame))
            end_index = min(end_index + self.batch_size, len(frame))
        return self.write_successed
    def check_metadata(self, columns, **kwargs):
        pass
    
    def create_table(self, table_name, columns, **kwargs):
        pass


class DataCommonWriter(AbstractEngineWriter):
    def __init__(self, output_config):
        super().__init__()
        self.request_timeout = 3000
        self.batch_size = 3000
        self.default_column = ['id', 'batch_no', 'ts']
        self.id_prefix = str(uuid.uuid4())
        try:
            if not output_config:
                raise Exception("OUTPUT DATASET INFO 为空")
            # table
            self.table_meta = output_config['tableMeta']
            self.logic_table_id = self.table_meta['tableId']
            self.train_batch = self.table_meta['trainBatch']
            self.table_name = self.table_meta['tableName']
            self.table_columns = self.table_meta['columns']

            # params
            self.params = output_config['params']
            self.output_addr = self.params['outputURL']
            self.datasource_id = self.params['outputSourceID']

            if len(self.train_batch) == 0:
                raise Exception("train batch 为空")
        except Exception as e:
            traceback.print_exc()
            raise OutputConfigException(e)

    def check(self, user_data) -> pd.DataFrame:
        if self.logic_table_id == "" and (self.table_name == "" or self.datasource_id == ""):
            raise OutputConfigException(
                "logic_table_id/table_name、datasource_id 不能都为空")

        # 表字段校验
        table_column = {c['code']: True for c in self.table_columns}
        user_column = [c['code'] for c in user_data['meta']]

        if not set(user_column).issubset(table_column):
            raise Exception(
                f'用户数据存在数据表中不存在字段，数据表字段为:{table_column.keys()}, 用户字段为:{user_column}')

        # 数据校验
        rows = user_data['result']
        df = pd.DataFrame(columns=user_column, data=rows)

        # 清除关键字
        for c in self.default_column:
            if c in user_column:
                df = df.drop(labels=c, axis=1)
                logger.warn(f"[Output] DataCommonWriter drop user column: {c}")

        return df

    def _push_to_source(self, df: pd.DataFrame):
        columns = self.default_column + df.columns.tolist()
        req = {
            "logicTableId": self.logic_table_id,
            "ytenantId": config.TENANT_ID,
            "datasourceId": self.datasource_id,
            "dataModel": {
                "tableName": self.table_name,
                "columnSchema": columns,
            }
        }
        total = df.shape[0]
        df.insert(0, 'id', self.id_prefix + df.index.astype(str))
        split_data = None
        for idx in range(0, total, self.batch_size):
            ts = str(datetime.datetime.now())
            end_idx = min(total, idx + self.batch_size)
            logger.info("[Output] DataCommonWriter push data index: %d to %d, total: %d",
                        idx, end_idx, total)
            split_data = df[idx:end_idx].copy()
            split_data.insert(1, 'batch_no', self.train_batch)
            split_data.insert(2, 'ts', ts)
            req['dataModel']['modelData'] = split_data.values.tolist()
            resp = iuap_request.post_json(
                url=self.output_addr, json=req,
                headers={'service-code': 'iuap-aip-console'},
                timeout=self.request_timeout,
                auth_type=iuap_request.AuthType.YHT)
            resp.raise_for_status()
            resp_content = resp.json
            logger.info("[Output] DataCommonWriter response: %s", resp_content)
            if not isinstance(resp_content, dict) or resp_content.get("code") == 1:
                req['dataModel']['modelData'] = "数据隐藏"
                raise OutputRequestException(
                    f"request: {json.dumps(req, ensure_ascii=False, indent=1)},\nresponse: {resp_content}")

        del split_data
        return True

    def _push_to_s3(self, df: pd.DataFrame):
        curkey = os.path.join(config.STORAGE_SERVICE_PATH,
                              self.train_batch, "output.csv")
        with tempfile.NamedTemporaryFile() as temp:
            filename = temp.name + '.csv'
            df.to_csv(filename, index=False)
            StorageService(curkey, "upload").upload(filename)
        return curkey

    def write(self, data):
        try:
            df = self.check(data)
        except Exception as e:
            traceback.print_exc()
            raise OutputCheckException(e)

        try:
            self._push_to_source(df)
        except Exception as e:
            logger.error(
                f"[Output]DataCommonWriter push data error:{e}, try to s3")
            curkey = self._push_to_s3(df)
            raise OutputRequestException(
                f"DataCommonWriter push data error:{e}, try to s3: {curkey}")
        return self.write_successed

    def check_metadata(self, columns, **kwargs):
        pass
    
    def create_table(self, table_name, columns, **kwargs):
        pass

class DataIwFactoryDrWriter(AbstractEngineWriter):
    '''
    数据写入数据工厂物化表
    1. 检验字段是否合规 check_metadata()
    2. 创建物化表 create_table()
    3. 写数据 write()
    '''
    def __init__(self, output_config):
        super().__init__()
        self.request_timeout = config.DATA_SOURCE_WRITE_TIMEOUT
        self.batch_size = config.DATA_SOURCE_WRITE_SIZE
        try:
            if not output_config or 'params' not in output_config:
                raise Exception("OUTPUT DATASET INFO 为空")

            # params
            self.params = output_config['params']
            self.output_addr = self.params.get('outputURL') # 输出数据到物化表接口
            self.datasource_id = self.params.get('outputSourceID') # 数据源id
            # 项目名称
            self.project_name = self.params.get('projectName', config.IWFACTORY_DR_PROJECT_NAME)
            self.metadata_url = self.params.get('metadataURL') # 校验建表字段是否合规接口，返回规范化字段
            self.create_url = self.params.get('createURL') # 创建物化表接口，返回表id，表名等数据

            user_url_map = {
                'std_metadata': self.metadata_url,  # 表元数据规范
                'physics_table': self.create_url,  # 物化建表
                'write_data': self.output_addr,  # 数据批量写入
            }

            self.iwfactory_dr_api = IwFactoryDrAPI(
                tenant_id=config.TENANT_ID,
                user_url_map=user_url_map
            )

        except Exception as e:
            traceback.print_exc()
            raise OutputConfigException(e)

    def create_table(self, table_name, columns, **kwargs):
        '''
        创建物化表
        :return:
        '''
        try:
            description = kwargs.get("description")
            result = self.iwfactory_dr_api.create_physics_table(self.datasource_id, self.project_name, table_name, description, columns)
            final_result = {
                "table_id": result.get("data", {}).get("uniqueNo"),
                "origin_result": result,
            }

        except Exception as e:
            traceback.print_exc()
            raise OutputRequestException(e)
        return final_result

    def _convert_columns(self, columns):
        '''
        原始格式为 {
            "code": "字段名",
            "type": "字段类型",
            "label": "字段描述"
        }
        转化后格式 {
                "columnEnglishName": "字段名",
                "columnDescribe": "字段描述",
                "columnType": "字段类型",
                "columnLength": "字符长度, INT=11, VARCHAR=255",
            }
        '''
        new_columns = []
        for column in columns:
            if not column.get("code") or not column.get("type"):
                raise OutputRequestException("字段名或字段类型不能为空")
            new_column = {
                "columnEnglishName": column.get("code", ""),
                "columnDescribe": column.get("label", ""),
                "columnType": column.get("type", ""),
                "columnLength": self._get_column_length(column.get("type", ""))
            }
            new_columns.append(new_column)
        return new_columns

    def _get_column_length(self, column_type):
        """
        根据字段类型返回对应的字符长度
        """
        type_upper = column_type.upper() if column_type else ""
        if "INT" in type_upper:
            return 11
        elif "VARCHAR" in type_upper or "STRING" in type_upper or "TEXT" in type_upper:
            return 255
        elif "FLOAT" in type_upper or "DOUBLE" in type_upper or "DECIMAL" in type_upper:
            return 20
        elif "DATE" in type_upper:
            return 10
        elif "DATETIME" in type_upper or "TIMESTAMP" in type_upper:
            return 19
        else:
            return 255

    def check_metadata(self, columns):
        f'''
        校验建表字段是否合规
        :return:
        '''
        try:
            new_metadata = self._convert_columns(columns)
            print(new_metadata)
            result = self.iwfactory_dr_api.std_metadata(self.datasource_id, new_metadata)
            final_result = {
                "columns": result.get("data", {}).get("columns"),
                "origin_result": result,
            }

        except Exception as e:
            traceback.print_exc()
            raise OutputRequestException(e)
        return final_result

    def write(self, data:pd.DataFrame, **kwargs):
        '''
        写数据到数据工厂物化表
        :param table_id:物化表id
        :param table_name:物化表名
        :param df:数据
        :return:
        '''
        try:
            logic_table_id, table_name = kwargs.get('table_id'), kwargs.get('table_name')
            if not logic_table_id or not table_name:
                raise Exception("table_id or table_name is empty")
            # 替换 Nan 为None
            data = data.replace({np.NAN: None})
            columns = data.columns.tolist()

            total = data.shape[0]
            split_data = None
            for idx in range(0, total, self.batch_size):
                end_idx = min(total, idx + self.batch_size)
                logger.info("[Output] DataAIWriter push data index: %d to %d, total: %d", idx, end_idx, total)
                split_data = data[idx:end_idx].copy()
                result = self.iwfactory_dr_api.write_data(self.datasource_id, table_name, columns, split_data.values.tolist(), logic_table_id)
                logger.info("[Output] DataAIWriter response: %s", result)
            del split_data
        except Exception as e:
            logger.error(
                f"[Output]DataAIWriter push data error:{e}, try to s3")
            curkey = self._push_to_s3(logic_table_id, data)
            raise OutputRequestException(
                f"DataAIWriter push data error:{e}, try to s3: {curkey}")
        return self.write_successed

    # def write(self, logic_table_id, table_name, table_columns, data):
    #     '''
    #     写数据到数据工厂物化表
    #     :param logic_table_id:物化表id
    #     :param table_name:物化表名
    #     :param table_columns:字段
    #     :param data:数据
    #     :return:
    #     '''
    #     try:
    #         result = self.iwfactory_dr_api.write_data(self.datasource_id, table_name, table_columns, data, logic_table_id)
    #     except Exception as e:
    #         traceback.print_exc()
    #         raise OutputRequestException(e)
    #     return result

    def _push_to_s3(self, logic_table_id, df: pd.DataFrame):
        curkey = os.path.join(config.STORAGE_SERVICE_PATH, logic_table_id, "output.csv")
        with tempfile.NamedTemporaryFile() as temp:
            filename = temp.name + '.csv'
            df.to_csv(filename, index=False)
            StorageService(curkey, "upload").upload(filename)
        return curkey

class DataSourceWriter(AbstractDataSourceWriter):
    def __init__(self, **kwargs):
        super().__init__()
        self.writer_info = kwargs
        self.writer_type = kwargs.get("writer_type")

        try:
            self.writer = self.__get_writer()
            self.table_columns = self.writer.table_columns
        except Exception as e:
            raise DataSourceWriterException(
                f"Data Output Initialization Error: {e}")

    def __get_writer(self) -> AbstractEngineWriter:
        if self.writer_type == DST.SQL:
            return SqlWriter()
        elif self.writer_type in (DST.IW_FACTORY_DATA, DST.INTELLIV):
            output_config = self.writer_info["output_config"]
            return DataCommonWriter(output_config)
        elif self.writer_type in [DST.IW_FACTORY_DR_DATA]:
            output_config = self.writer_info["output_config"]
            return DataIwFactoryDrWriter(output_config)
        else:
            return UserWriter()

    def write(self, data, **kwargs):
        try:
            return self.writer.write(data=data, **kwargs)
        except Exception as e:
            raise DataSourceWriterException(f"Data Output Process error: {e}")


    def check_metadata(self, columns, **kwargs):
        try:
            return self.writer.check_metadata(columns=columns, **kwargs)
        except Exception as e:
            raise DataSourceWriterException(f"Data check metadata Process error: {e}")
    
    def create_table(self, table_name, columns, **kwargs):
        try:
            return self.writer.create_table(table_name=table_name, columns=columns, **kwargs)
        except Exception as e:
            raise DataSourceWriterException(f"Data create table Process error: {e}")


"""
Author: Hexu
Date: 2022-05-05 12:07:16
LastEditors: Hexu
LastEditTime: 2024/9/23 09:56
FileName: __init__.py
Description:
"""
import os

from intelliw.utils.kafka.client import KafkaClient, handle_message, delivery_callback

from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


class KafkaAuthTypeEnum:
    PLAIN = 'plain'
    SCRAM = 'scram'
    SCRAM_SSL = 'scram_ssl'
    SSL = 'ssl'


AUTH_TYPE_CONFIG = {
    KafkaAuthTypeEnum.PLAIN: {
        'security_protocol': 'SASL_PLAINTEXT',
        'sasl_mechanism': 'PLAIN'
    },
    KafkaAuthTypeEnum.SCRAM: {
        'security_protocol': 'SASL_PLAINTEXT',
        'sasl_mechanism': 'SCRAM-SHA-256'
    },
    KafkaAuthTypeEnum.SCRAM_SSL: {
        'security_protocol': 'SASL_SSL',
        'sasl_mechanism': 'SCRAM-SHA-256'
    },
    KafkaAuthTypeEnum.SSL: {
        'security_protocol': 'SSL',
        'sasl_mechanism': None
    },
}


def _get_args(brokers: str, username: str, password: str, auth_type: str = None):
    """
    获取 Kafka 连接参数，优先使用传入参数，若未提供则从环境变量中获取
    """
    args = {
        'brokers': brokers or os.environ.get('kafka.cluster.hosts', ''),
    }

    if not args['brokers']:
        raise KeyError('Brokers information is missing')

    kafka_auth_type = auth_type or os.environ.get('kafka.authentication.type', '')

    if kafka_auth_type:
        auth_type_lower = kafka_auth_type.lower()
        if auth_type_lower in AUTH_TYPE_CONFIG:
            config = AUTH_TYPE_CONFIG[auth_type_lower]
            args['security_protocol'] = config['security_protocol']
            args['sasl_mechanism'] = config['sasl_mechanism']
            args['auth_type'] = kafka_auth_type
        else:
            logger.warning(f"Unknown kafka.authentication.type: {kafka_auth_type}, will use default")
    else:
        kafka_auth_enable = os.environ.get('kafka.authentication.enable', '')
        if kafka_auth_enable.lower() == 'true':
            args['username'] = username or os.environ.get('kafka.username', '')
            args['password'] = password or os.environ.get('kafka.password', '')

    if kafka_auth_type in (KafkaAuthTypeEnum.PLAIN, KafkaAuthTypeEnum.SCRAM, KafkaAuthTypeEnum.SCRAM_SSL):
        args['username'] = username or os.environ.get('kafka.username', '')
        args['password'] = password or os.environ.get('kafka.password', '')
    elif not kafka_auth_type:
        kafka_auth_enable = os.environ.get('kafka.authentication.enable', '')
        if kafka_auth_enable.lower() == 'true':
            args['username'] = username or os.environ.get('kafka.username', '')
            args['password'] = password or os.environ.get('kafka.password', '')
            args['security_protocol'] = 'SASL_PLAINTEXT'
            args['sasl_mechanism'] = 'PLAIN'

    return args


def get_client(brokers: str = None, group_id: str = None, username: str = None, password: str = None,
               producer_config: dict = None, consumer_config: dict = None,
               security_protocol: str = None, sasl_mechanism: str = None,
               auth_type: str = None,
               key_serializer=None, value_serializer=None, ssl_ca_location: str = None,
               ssl_endpoint_identification_algorithm: str = 'none'):
    """
    创建 Kafka 客户端实例
    :param brokers: Kafka 集群地址
    :param group_id: 消费者分组 ID
    :param username: SASL 验证用的用户名
    :param password: SASL 验证用的密码
    :param producer_config: 生产者配置
    :param consumer_config: 消费者配置
    :param security_protocol: 安全协议，若设置 auth_type 则自动推断
    :param sasl_mechanism: SASL 认证机制，若设置 auth_type 则自动推断
    :param auth_type: 认证类型 (plain, scram, scram_ssl, ssl)
    :param key_serializer: 用于序列化 key 的函数
    :param value_serializer: 用于序列化 value 的函数
    :param ssl_ca_location: SSL 证书路径，若不指定则默认为 /etc/ssl/certs/ca-certificates.crt
    :param ssl_endpoint_identification_algorithm: SSL 端点标识算法，默认为 'none'，可设为 'hostname' 启用主机名验证
    :return: KafkaClient 实例或 None（如果获取参数失败）
    """

    try:
        args = _get_args(brokers, username, password, auth_type)

        if ssl_ca_location is None:
            ssl_ca_location = os.environ.get('kafka.ssl.ca.location', '/etc/ssl/certs/ca-certificates.crt')

        kafka_client = KafkaClient(
            brokers=args['brokers'],
            group_id=group_id,
            username=args.get('username'),
            password=args.get('password'),
            producer_config=producer_config,
            consumer_config=consumer_config,
            security_protocol=security_protocol or args.get('security_protocol'),
            sasl_mechanism=sasl_mechanism or args.get('sasl_mechanism'),
            auth_type=auth_type or args.get('auth_type'),
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            ssl_ca_location=ssl_ca_location,
            ssl_endpoint_identification_algorithm=ssl_endpoint_identification_algorithm
        )
        return kafka_client

    except KeyError as e:
        logger.error(f"Kafka args error: {e}")
        return None



import json
import os
from typing import Tuple

from intelliw.config import config


def set_dataset_config(dataset_id: str = None, dataset_ratio: list = None,
                       local_csv: Tuple[str, list] = None,
                       local_corpus: str = None):
    """通过ai工作坊接口设置数据集所需的环境变量

    Args:
        dataset_id (str): 数据集id, 从ai工作坊页面上获取
        dataset_ratio(list): 数据机训练集，验证集，测试集比例， 0.7:0.3:0.1
        local_csv(str, list): 本地csv文件位置，支持多数据集合
        local_corpus(str): 本地nlp语料文件位置

    所需环境变量:
        SOURCE_TYPE = 3
        INPUT_MODEL_ID = ''
        INPUT_DATA_SOURCE_ID = ''
        INPUT_DATA_SOURCE_TRAIN_TYPE = 2
        DATA_SOURCE_ADDRESS = ''

        if SOURCE_TYPE = 1
            NEED: DATA_SOURCE_ADDRESS
        if SOURCE_TYPE = 2
            NEED: INPUT_ADDR, INPUT_GETROW_ADDR, INPUT_MODEL_ID
        if SOURCE_TYPE = 4
            NEED: INPUT_ADDR, INPUT_GETROW_ADDR, INPUT_DATA_SOURCE_ID, INPUT_DATA_SOURCE_TRAIN_TYPE
        if SOURCE_TYPE = 5
            NEED: INPUT_ADDR, INPUT_GETROW_ADDR, INPUT_DATA_SOURCE_ID
        if SOURCE_TYPE = 21
            NEED: INPUT_ADDR, INPUT_GETROW_ADDR, INPUT_DATA_SOURCE_ID,
                  INPUT_DATA_SOURCE_TRAIN_TYPE, NLP_CORPORA_INPUT_TYPE
    """

    if dataset_ratio is None:
        dataset_ratio = [0.7, 0.3, 0]

    if not any([local_csv, local_corpus, dataset_id]):
        config.SOURCE_TYPE = 0
        return
    config.SOURCE_TYPE = -1

    if not isinstance(local_csv, (list, str)):
        local_csv = None

    if local_csv:
        _csv_file_check = all(os.path.exists(c) for c in local_csv) if isinstance(
            local_csv, list) else os.path.exists(local_csv)
    else:
        _csv_file_check = False
    _nlp_corpus_check = os.path.exists(local_corpus) if local_corpus else False

    if not any([dataset_id, _csv_file_check, _nlp_corpus_check]):
        raise ValueError(
            "数据集错误,请通过--csv设置本地csv文件 或 -C设置本地nlp语料文件夹, 或者--dataset设置在线数据集")

    config.TRAIN_DATASET_RATIO = dataset_ratio[0]
    config.VALID_DATASET_RATIO = dataset_ratio[1]
    config.TEST_DATASET_RATIO = dataset_ratio[2]

    dataset_list = []
    if not dataset_id:
        if local_csv:
            if isinstance(local_csv, str):
                dataset_list = [{"SOURCE_TYPE": 3, "CSV_PATH": local_csv}]
            elif isinstance(local_csv, list):
                dataset_list = [{"SOURCE_TYPE": 3, "CSV_PATH": c}
                                for c in local_csv]
        elif local_corpus:
            dataset_info = {"SOURCE_TYPE": 21, "NLP_CORPORA_PATH": local_corpus,
                            "NLP_CORPORA_INPUT_TYPE": "local", "INPUT_DATA_SOURCE_ID": "",
                            "INPUT_DATA_SOURCE_TRAIN_TYPE": config.INPUT_DATA_SOURCE_TRAIN_TYPE}
            dataset_list.append(dataset_info)
    else:
        config.update_by_env()
        from intelliw.utils import iuap_request
        # Shit, shit
        if os.environ.get('useGpaas') == 'true':
            dataset_url = f"{os.environ.get('domain.url')}{os.environ.get('DATASET_BY_ID_ADDRESS')}"
        else:
            dataset_url = os.environ.get('DATASET_BY_ID_ADDRESS')
        for i in dataset_id.split(","):
            dataset_info = {}
            resp = iuap_request.get(dataset_url, params={"dataSetId": i})
            resp.raise_for_status()
            body = resp.json
            if body['status'] == 0:
                raise ConnectionAbortedError(
                    f"get dataset info response: {body}")
            result = body['data']
            source = result["SOURCE_TYPE"]
            if result.get("TENANT_ID"):
                dataset_info["TENANT_ID"] = result.get("TENANT_ID")
                config.TENANT_ID = result.get("TENANT_ID")
            if source == 1:
                dataset_info["DATA_SOURCE_ADDRESS"] = result["DATA_SOURCE_ADDRESS"]
            else:
                dataset_info["INPUT_ADDR"] = result["INPUT_ADDR"]
                dataset_info["INPUT_GETROW_ADDR"] = result["INPUT_GETROW_ADDR"]
                if source == 2:
                    mid = result["INPUT_MODEL_ID"]
                    dataset_info["INPUT_GETROW_ADDR"] = result["INPUT_GETROW_ADDR"]
                    dataset_info["INPUT_MODEL_ID"] = mid
                elif source == 4:
                    dataset_info["INPUT_DATA_SOURCE_ID"] = result["INPUT_DATA_SOURCE_ID"]
                elif source == 5:
                    dataset_info["INPUT_DATA_SOURCE_ID"] = result["INPUT_DATA_SOURCE_ID"]
                    dataset_info["INPUT_DATA_META_ADDR"] = result["INPUT_DATA_META_ADDR"]
                elif source == 9:
                    dataset_info["INPUT_ADDR"] = result["INPUT_ADDR"]
                    dataset_info["INPUT_MODEL_ID"] = result["INPUT_MODEL_ID"]
                    dataset_info["USER_ID"] = os.environ.get("USER_ID")
                    dataset_info["DATA_FILTER_CONDITION"] = os.environ.get("DATA_FILTER_CONDITION")
                elif source == 21:
                    dataset_info["INPUT_DATA_SOURCE_ID"] = result["INPUT_DATA_SOURCE_ID"]
                    dataset_info["NLP_CORPORA_INPUT_TYPE"] = result["NLP_CORPORA_INPUT_TYPE"]
                elif source == 24:
                    dataset_info["INPUT_DATA_SOURCE_ID"] = result["INPUT_DATA_SOURCE_ID"]
                    dataset_info["FINE_TUNING_SET_TYPE"] = dataset_info["FINE_TUNING_SET_TYPE"]
            dataset_info["SOURCE_TYPE"] = source
            dataset_list.append(dataset_info)
    config.DATASET_INFO = json.dumps(dataset_list, ensure_ascii=False)



import threading
from typing import Type, Any, Dict


def singleton(cls: Type) -> Type:
    """
    单例模式装饰器
    确保被装饰的类只有一个实例
    """
    instances: Dict[Type, Any] = {}
    lock = threading.RLock()  # 使用可重入锁

    def get_instance(*args: Any, **kwargs: Any) -> Any:
        if cls not in instances:
            with lock:
                if cls not in instances:
                    instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/2/22 16:45
# @Author  : HEXU
# @File    : test.py
# @Description :
import os
import asyncio
from datetime import datetime

from intelliw.utils.redis import get_client, get_async_client

os.environ['iuap-aip-console-redis_REDIS_TYPE'] = 'cluster'
os.environ['iuap-aip-console-redis_REDIS_URL'] = '172.20.32.142:6379,172.20.32.195:6379,172.20.32.16:6379'
os.environ['iuap-aip-console-redis_REDIS_PASSWORD'] = 'Vo14OCmocLJAC9Nd'
os.environ['iuap-aip-console-redis_REDIS_SENTINEL_PASSWORD'] = ''
os.environ['iuap-aip-console-redis_REDIS_MASTER_NAME'] = ''
os.environ['iuap-aip-console-redis_REDIS_DATABASE'] = ''
os.environ['iuap-aip-console-redis_REDIS_POOL_SIZE'] = '20'


def sync_client():
    rs = get_client()

    data_key = "intelliw:N:test"
    result_tag = rs.set(data_key, str(datetime.now()))
    print("sync_client set", result_tag)
    value = rs.get(data_key)
    print("sync_client get", value)


async def async_client():
    rs = get_async_client()

    data_key = "intelliw:N:test"
    result_tag = await rs.set(data_key, str(datetime.now()))
    print("async_client set", result_tag)
    value = await rs.get(data_key)
    print("async_client set", value)


if __name__ == '__main__':
    sync_client()
    asyncio.run(async_client())



def prepare_model_params(model_data):
    """
    Prepare model parameters for processing.
    
    Args:
        model_data: Raw model data to be processed
        
    Returns:
        Processed model data
    """
    # For now, just return the model data as is
    # In a real implementation, this would process the model data
    return model_data


# type: ignore
'''
Author: hexu
Date: 2022-02-20 10:49:18
LastEditTime: 2023-03-16 17:16:15
LastEditors: Hexu
Description: 按列的特征处理
FilePath: /iw-algo-fx/intelliw/functions/feature_process.py
'''
from copy import deepcopy
import json
import random
from intelliw.utils import message
from intelliw.utils.util import get_json_encoder
from collections.abc import Iterable
from intelliw.config import cfg_parser
from intelliw.config import config as fw_config
from intelliw.utils.logger import _get_framework_logger
from intelliw.utils.global_val import gl
from intelliw.utils.exception import FeatureProcessException
from intelliw.utils.exception import DatasetException
from intelliw.utils.spark_process import Engine

try:
    import numpy as np
    from sklearn.cluster import KMeans
except ImportError:
    raise ImportError(
        "\033[31mIf use feature process, you need: pip install scikit-learn\033[0m")

logger = _get_framework_logger()

pd = Engine()

COl_PREFIX = "col_"

stage_feature_map = {
    "pre-train": "train",
    "post-train": "train",
    "pre-predict": "infer",
    "post-predict": "infer",
    "infer": "infer",
    "train": "train",
}

stage_infer_map = {
    "pre-train": "pre-predict",
    "post-train": "post-predict",
}


# injection_alg_yaml 新的配置注入到旧的配置里
# yaml_path 配置文件位置
# 能到这步key都是存在的，不用担心没有key


def injection_alg_yaml(yaml_path, val, stage, type):
    cfg = cfg_parser.load_single_config(yaml_path)
    value = [
        {
            "key": "feature_process_param",
            "desc": "特征处理参数",
            "name": "特征处理参数",
            "val": val,
            "option": {"defaultValue": "", "type": "string"}
        }
    ]
    if type == "algorithm":
        transforms = cfg["AlgorithmInformation"]["algorithm"]["transforms"]
    else:
        transforms = cfg["Model"]["transforms"]

    for t in transforms:
        if t["type"] == stage:
            for f in t["functions"]:
                if f["key"] == "feature_process":
                    f["parameters"] = value
                    break
            else:
                t["functions"].append(
                    {"key": "feature_process", "parameters": value})
            break
    else:
        transforms.append({"type": stage, "functions": [
            {"key": "feature_process", "parameters": value}]})
    return cfg


def feature_process(data, config: list, stage: str) -> Iterable:
    """特征工程处理

    Args:
        data (object): 需要处理的数据
        config (list): 特征值配置
        stage (str): 处理类型 pre-train/post-train/pre-predict/post-predict
        _type (str): 数据集类型 验证集/训练集  如果胃验证集应按照推理的方式处理

    Returns:
        Iterable: _description_
    """
    alg_yaml_file, model_yaml_file = gl.alg_yaml_file, gl.model_yaml_file
    if not config and not alg_yaml_file and not model_yaml_file:
        return data

    logger.info("feature process start")
    processor_mode = stage_feature_map[stage]

    if processor_mode == "infer":
        # if "data" not in data:
        #     raise KeyError("如果使用特征处理功能, 推理请求的输入必须放在data参数下")
        # infer_data = data["data"]
        # df = pd.DataFrame(infer_data)
        # _, col = df.shape
        # if len(config) != col:
        #     raise KeyError(f"输入特征数量不正确, 输入特征数量{col}, 模型需要特征数量{len(config)}")
        # fp = FeatureProcessor(config, processor_mode)
        # result, _, _, _ = fp.do_process(df, fit=False)
        # data["data"] = result.values.tolist()
        # data["original_data"] = infer_data
        # logger.info("feature process end")
        return data
    else:
        # 获取所有数据
        alldata = data
        result = alldata.get("result")
        # if not result:
        #     return data
        #
        # column_meta = [col.get("column_code", col["column_name"])
        #                for col in config]
        # df = pd.DataFrame(result)
        # shape = df.shape
        #
        # # 特征处理可能会修改目标列的属性，所以需要先保存起来
        # target_metadata = gl.get("target_metadata")
        # target_data = None
        # if target_metadata and len(target_metadata) > 0:
        #     target_cols = [m['target_col'] for m in target_metadata]
        #
        #     # 1 如果出现目标为空，做筛选
        #     df = df.dropna(subset=target_cols)
        #     if df.shape[0] / shape[0] < 0.95:
        #         raise DatasetException("目标列空值过多, 请检查数据")
        #
        #     # 2 提取目标列
        #     target_data = df.iloc[:, target_cols]
        #     # 检查目标列是否要做特征工程，如果是删除目标列特征配置
        #     for idx, c in enumerate(deepcopy(config)):
        #         if c["col"] in target_cols:
        #             config.pop(idx)
        #
        # fp = FeatureProcessor(config, processor_mode)
        # result, category_cols, new_config, col_map_df = fp.do_process(
        #     df, fit=True)
        #
        # # 将目标列加入数据
        # if target_data is not None:
        #     target_names = [m['target_name'] for m in target_metadata]
        #     target_data.columns = target_names
        #     result = pd.concat([result, target_data], axis=1)
        #     column_meta.extend(target_names)
        #     target_cols = [{"col": result.columns.get_loc(
        #         name), "column_name": name} for name in target_names]
        #     gl.target_cols = target_cols
        #     # TODO 暂时兼容后面删除
        #     gl.final_target_col = target_cols[-1]["col"]
        #
        # # 原始列名
        # alldata["meta"] = [{'code': i} for i in result.columns.tolist()]
        # alldata["result"] = result.values.tolist()
        # gl.set_dict({"category_cols": category_cols,
        #              "column_relation_df": col_map_df, "column_meta": column_meta})
        # logger.info("feature process end")
        #
        # # 如果是训练需要写入配置
        # gl.get('feature_process')[stage] = new_config
        # new_config = json.dumps(
        #     new_config, ensure_ascii=False, cls=get_json_encoder())
        #
        # if not fw_config.is_server_mode():
        #     # 本地测试需要反写入配置文件
        #     model_yaml = injection_alg_yaml(
        #         model_yaml_file, new_config, stage_infer_map[stage], "model")
        #     cfg_parser.dump_config(model_yaml, model_yaml_file)
        #     alg_yaml = injection_alg_yaml(
        #         alg_yaml_file, new_config, stage_infer_map[stage], "algorithm")
        #     cfg_parser.dump_config(alg_yaml, alg_yaml_file)
        #     logger.info("injection feature process config to yaml")
        # elif gl.recorder is not None:
        #     # 线上的上报到AI工作坊
        #     reporter_info = {"modelInstanceId": fw_config.INSTANCE_ID,
        #                      "tenantId": fw_config.TENANT_ID,
        #                      "valuationResult": {stage_infer_map[stage]: new_config}}
        #     gl.recorder.report(message.CommonResponse(
        #         200, 'feature_config', 'feature_config_upload', reporter_info))
        #     logger.info("report feature process config to ai-console")
        return alldata


class FeatureProcessor:
    def __init__(self, config_list, mode):
        self.config_list = config_list
        self.mode = mode  # train or infer
        self.l_Features = []
        self.col_names = []
        self.col_nums = []
        self.col_map_df = {'origin_num': [],
                           'origin_name': [], 'final_num': [], 'final_name': []}

        if self.mode == 'train':
            for col_config in config_list:
                idx = col_config['col']
                col_name = col_config.get(
                    "column_code", col_config["column_name"])
                self.col_names.append(col_name)
                self.col_nums.append(idx)
                if col_config['type'].lower() in ['int', 'float', 'double', 'number']:
                    feat = NumericFeature(idx, col_config, self.mode)
                elif col_config['type'].lower() in ['string', 'category']:
                    feat = CategoryFeature(idx, col_config, self.mode)
                else:
                    feat = UnknownFeature(idx, col_config, self.mode)
                self.l_Features.append(feat)

        elif self.mode == 'infer':
            for idx, col_config in enumerate(config_list):
                col_name = col_config.get(
                    "column_code", col_config["column_name"])
                self.col_names.append(col_name)
                self.col_nums.append(idx)
                if col_config['type'].lower() in ['int', 'float', 'double', 'number']:
                    feat = NumericFeature(idx, col_config, self.mode)
                elif col_config['type'].lower() in ['string', 'category']:
                    feat = CategoryFeature(idx, col_config, self.mode)
                else:
                    feat = UnknownFeature(idx, col_config, self.mode)
                self.l_Features.append(feat)

    def do_process(self, data_frame, fit):
        '''
        :param data_frame:
        :return: when self.mode == 'train', processor will fit params with input data, if you just want to do data
                 process without change params, use set_mode('infer') before do_process!

                 df: the output dataframe transformed from data_frame.

                 category_cols: the col_nums of columns which are in category type.

                 config_list: the processing methods used in missing check,miss repairing, outlier check, outlier
                              repairing for each column.

                 col_map_df: the map between the input columns and output columns,ie.:

                                        origin_num origin_name  final_num final_name
                                0           0          id          0          0
                                1           1        name          1          1
                                2           2        math          2          2
                                3           3       level          3          3
                                4           5        ABCD          4        5_0
                                5           5        ABCD          5        5_1
                                6           5        ABCD          6        5_2
                                7           5        ABCD          7        5_3
                                8           5        ABCD          8        5_4
                                9           6        rand          9          6
                            Warning: if column is droped, it will not exist in col_map_df!
        '''

        df = pd.DataFrame()
        self.category_cols = []
        self.col_map_df = {'origin_num': [],
                           'origin_name': [], 'final_num': [], 'final_name': []}

        if self.mode == 'train':
            if fit:
                self.__set_fit(True)
            else:
                self.__set_fit(False)
        else:
            self.__set_fit(False)

        for i, feat in enumerate(self.l_Features):
            logger.info(
                f"Feature Processing: there are {len(self.l_Features)} columns, working on column {i + 1}")
            try:
                col_num = df.shape[1]
                tmp_frame, res_params = feat.do_process(data_frame)
                shape = tmp_frame.shape
                if len(shape) > 1:
                    if shape[1] > 0:  # shape[1]为0 时，则该列可能被drop掉了
                        self.col_map_df['origin_num'].extend(
                            [self.col_nums[i]] * shape[1])
                        self.col_map_df['final_num'].extend(
                            range(col_num, col_num + shape[1]))
                        self.col_map_df['origin_name'].extend(
                            [self.col_names[i]] * shape[1])
                        self.col_map_df['final_name'].extend(tmp_frame.columns)
                else:
                    self.col_map_df['origin_num'].append(self.col_nums[i])
                    self.col_map_df['final_num'].append(col_num)
                    self.col_map_df['origin_name'].append(self.col_names[i])
                    self.col_map_df['final_name'].extend(tmp_frame.columns)

                if isinstance(feat, CategoryFeature):
                    if len(shape) > 1:
                        self.category_cols.extend(
                            range(col_num, col_num + shape[1]))
                    else:
                        self.category_cols.append(col_num)
                cols = list(df.columns)
                cols.extend(tmp_frame.columns)
                df = pd.concat([df, tmp_frame], axis=1)
                df.columns = cols
                if fit and self.mode == 'train':
                    self.config_list[i]['func'] = res_params
            except Exception as e:
                import traceback
                err_stack = traceback.format_exc()
                raise Exception(
                    f"origin_col_num:{feat.index}, final_num:{col_num}, order of the column:{i}, stack:\n {err_stack}")

        self.__set_fit(False)
        self.col_map_df = pd.DataFrame(self.col_map_df)
        return df, self.category_cols, self.config_list, self.col_map_df

    def __set_fit(self, fit):
        if fit:
            for feat in self.l_Features:
                feat.mode = 'train'
        else:
            for feat in self.l_Features:
                feat.mode = 'infer'


class Feature:
    def __init__(self, index, config, mode, **kwargs):
        '''
        :param index:
        :param config:
        :param mode:
        :param auto_process: when mode == 'train' and auto_process == True the recommended func will be used, and params
                             will write back to the feature_engineer_config!
        :param auto_check_repair: when mode == 'train' and auto_check_repair == True, the recommended check_and_repair
                             method will be used, and params will write back to the feature_engineer_config!
        '''
        self.index = index
        self.mode = mode
        self.type = config['type']
        self.check_repair_config = config['check_repair']
        self.feature_engineer_config = config['func']
        self.fe_name2func = {}
        self.auto_check_repair = True if config.get(
            'check_repair_auto', '0') == '1' else False
        self.auto_process = True if config.get(
            'func_auto', '0') == '1' else False
        self.group_cols = config.get('group_cols', None)

    def do_process(self, data_frame):
        pass


class NumericFeature(Feature):
    def __init__(self, index, config, mode, **kwargs):
        super(NumericFeature, self).__init__(index, config, mode, **kwargs)
        self.fe_name2func = {'bucket': bucket,
                             'max_abs_scale': max_abs_scale,
                             'min_max_scale': min_max_scale,
                             'standard_scale': standard_scale,
                             'log_transform': log_transform,
                             'quantile_transform': quantile_transform,
                             'square': square,
                             'translation': translation,
                             'robust_scale': robust_scale,
                             'drop': drop,
                             'interval_bucket': interval_bucket,
                             'kmeans_bucket': kmeans_bucket
                             }

    def recommend_miss_repair(self, feat_df):
        if (feat_df == feat_df.mode()[0]).sum() / len(feat_df) >= 0.75:
            return '4'
        if feat_df.std() / feat_df.mean() <= 1:
            return '1'
        return '5'

    def recommend_outlier_check(self, feat_df):
        mean = feat_df.mean()
        std = feat_df.std()
        low_sigma = mean - 3 * std
        high_sigma = mean + 3 * std
        count_sigma = len(feat_df[feat_df < low_sigma]) + \
                      len(feat_df[feat_df > high_sigma])

        low_boxplot = feat_df.quantile(
            0.25) - 1.5 * (feat_df.quantile(0.75) - feat_df.quantile(0.25))
        high_boxplot = feat_df.quantile(
            0.75) + 1.5 * (feat_df.quantile(0.75) - feat_df.quantile(0.25))
        count_boxplot = len(feat_df[feat_df < low_boxplot]) + \
                        len(feat_df[feat_df > high_boxplot])

        if count_boxplot / len(feat_df) > 0.1 and count_sigma / len(feat_df) > 0.1:
            self.check_repair_config['outlier_check'] = '3'
            return 0, 0
        if count_sigma < count_boxplot:
            self.check_repair_config['outlier_check'] = '1'
            return low_sigma, high_sigma
        self.check_repair_config['outlier_check'] = '2'
        return low_boxplot, high_boxplot

    def recommend_outlier_repair(self, feat_df, low, high):
        if len(feat_df) < 1000 or (low == 0 and high == 0):
            return '3'
        return '1'

    def recommend_process(self, feat_df, low, high):
        # kurt = feat_df.kurt()
        coeff = feat_df.std() / feat_df.mean()
        skew = feat_df.skew()

        if coeff >= 5 and skew > 0:
            return [{'name': 'translation', 'params': {}},
                    {'name': 'log_transform', 'params': {}}]
        # if prop of outliers > 0.1
        if low == 0 and high == 0:
            return [{'name': 'robust_scale', 'params': {}}]
        # if skew == 0 or kurt < 5
        return [{'name': 'min_max_scale', 'params': {}}]

    def do_process(self, data_frame):
        feat_df = data_frame[data_frame.columns[self.index]]

        # check and repair
        if self.mode == 'train':
            # missing repair
            feat_df.replace([np.inf, -np.inf, 'None', 'null',
                             '', ' '], np.nan)
            feat_df = pd.to_numeric(feat_df, errors='coerce')
            if self.auto_check_repair:
                self.check_repair_config['miss_repair'] = self.recommend_miss_repair(
                    feat_df)

            if 'miss_repair' not in self.check_repair_config:
                raise FeatureProcessException(
                    f"origin_col_num:{self.index}, miss_repair config missing!")

            if not (self.check_repair_config['miss_repair'] in ['1', '2', '3', '4', '5', '6'] or type(
                    self.check_repair_config['miss_repair']) == dict):
                raise FeatureProcessException(
                    f"origin_col_num:{self.index}, miss_repair must be in ['1', '2', '3', '4', '5', '6'], or a dict")
                # 针对“7.分位数范围内随机填充”这种需要传参数的miss_repair 方法,用dict,待出通用方案,todo

            self.check_repair_config['index'] = self.index
            self.check_repair_config['group_cols'] = self.group_cols
            self.check_repair_config['data_frame'] = data_frame
            self.check_repair_config['type'] = self.type
            feat_df = numeric_miss_check_repair(
                self.check_repair_config, feat_df, self.mode)
            del self.check_repair_config['data_frame']

            # if feature has the same value, drop it
            if feat_df.nunique() == 1 and self.auto_process:
                self.feature_engineer_config = [{'name': 'drop', 'params': {}}]
                return pd.DataFrame(), self.feature_engineer_config

            # outlier check
            if self.auto_check_repair:
                low, high = self.recommend_outlier_check(feat_df)
            else:
                low, high = numeric_outlier_check(
                    self.check_repair_config, feat_df, self.mode)

            # outlier repair
            if self.auto_check_repair:
                self.check_repair_config['outlier_repair'] = self.recommend_outlier_repair(
                    feat_df, low, high)

            feat_df = numeric_outlier_repair(
                self.check_repair_config, feat_df, self.mode)

            if self.auto_process:
                self.feature_engineer_config = self.recommend_process(
                    feat_df, low, high)
        else:
            feat_df = pd.to_numeric(feat_df, errors='coerce')

        # feature engineering
        for func in self.feature_engineer_config:
            feat_df, params = self.fe_name2func[func['name']](
                func['params'], feat_df, self.mode)
            if self.mode == 'train':
                func['params'] = params

        cols = []
        if len(feat_df.shape) == 1:
            cols = [f'{self.index}_0']

        elif len(feat_df.shape) == 2:
            for i in range(feat_df.shape[1]):
                cols.append(f'{self.index}_{i}')

        feat_df.columns = cols
        return feat_df, self.feature_engineer_config


class CategoryFeature(Feature):
    def __init__(self, index, config, mode, **kwargs):
        super(CategoryFeature, self).__init__(index, config, mode, **kwargs)

        self.fe_name2func = {'one_hot': one_hot,
                             'ordinal': ordinal,
                             'drop': drop,
                             'hash': hash_feature
                             }

    def recommend_miss_repair(self, feat_df):
        if (feat_df == feat_df.mode()[0]).sum() / len(feat_df) >= 0.75:
            return '4'
        return '5'

    def recommend_process(self, feat_df):
        if feat_df.nunique() > len(feat_df) * 0.9:
            return [{'name': 'hash', 'params': {}}]
        if feat_df.nunique() <= 100:
            return [{'name': 'one_hot', 'params': {}}]
        return [{'name': 'ordinal', 'params': {}}]

    def do_process(self, data_frame):
        feat_df = data_frame[data_frame.columns[self.index]]
        # check and repair
        if self.mode == 'train':
            if self.auto_check_repair:
                self.check_repair_config['miss_repair'] = self.recommend_miss_repair(
                    feat_df)
            # missing repair
            if not ('miss_repair' in self.check_repair_config):
                raise FeatureProcessException(
                    f"origin_col_num:{self.index}, miss_repair config missing!")
            if not (self.check_repair_config['miss_repair'] in ['4', '5', '6']):
                raise FeatureProcessException(
                    f"origin_col_num:{self.index}, miss_repair must in ['4', '5', '6']!")

            self.check_repair_config['index'] = self.index
            self.check_repair_config['group_cols'] = self.group_cols
            self.check_repair_config['data_frame'] = data_frame
            self.check_repair_config['type'] = self.type
            feat_df = category_miss_check_repair(
                self.check_repair_config, feat_df, self.mode)
            del self.check_repair_config['data_frame']

            # if feature has the same value, drop it
            if feat_df.nunique() == 1 and self.auto_process:
                self.feature_engineer_config = [{'name': 'drop', 'params': {}}]
                return pd.DataFrame(), self.feature_engineer_config

        if self.mode == 'train' and self.auto_process:
            self.feature_engineer_config = self.recommend_process(feat_df)

        # feature engineering
        for func in self.feature_engineer_config:
            # if func['name'] in self.fe_name2func:
            feat_df, params = self.fe_name2func[func['name']](
                func['params'], feat_df, self.mode)
            if self.mode == 'train':
                func['params'] = params

        cols = []
        if len(feat_df.shape) == 1:
            cols = [f'{self.index}_0']

        elif len(feat_df.shape) == 2:
            for i in range(feat_df.shape[1]):
                cols.append(f'{self.index}_{i}')

        feat_df.columns = cols
        return feat_df, self.feature_engineer_config


def numeric_miss_check_repair(check_repair_config, col_frame, mode):
    group_cols = check_repair_config['group_cols']
    index = check_repair_config['index']
    data_frame = check_repair_config['data_frame']

    if check_repair_config['miss_repair'] == '1':
        col_frame = col_frame.fillna(col_frame.mean())
    elif check_repair_config['miss_repair'] == '2':
        col_frame = col_frame.fillna(0)
    elif check_repair_config['miss_repair'] == '3':
        col_frame = col_frame.fillna(col_frame.quantile(0.5))
    elif check_repair_config['miss_repair'] == '4':
        col_frame = col_frame.fillna(col_frame.mode()[0])
    elif check_repair_config['miss_repair'] == '6':
        if not (group_cols is not None and type(group_cols) == list and len(group_cols) != 0):
            raise FeatureProcessException(
                f"origin_col_num:{index}, when miss_repair='6', group_cols can't be empty")
        groups = data_frame.groupby(group_cols, observed=True).groups
        for group, idxs in groups.items():
            df_tmp = col_frame.loc[idxs].copy()
            if len(df_tmp[~df_tmp.isnull()]) == 0:
                fill_item = col_frame.mode()[0]  # 如果类内全是null，则用全局替代
            else:
                fill_item = df_tmp[~df_tmp.isnull()].mode()[0]
            col_frame.loc[idxs] = col_frame.loc[idxs].fillna(fill_item)
    elif type(check_repair_config['miss_repair']) == dict:
        if check_repair_config['miss_repair']['repair_type'] == '7':
            if not ('quantile_low' in check_repair_config['miss_repair'] and 'quantile_high' in check_repair_config[
                'miss_repair']):
                raise FeatureProcessException(
                    f"origin_col_num:{index}, when miss_repair='7', quantile_low and  quantile_high must be set")
            low = check_repair_config['miss_repair']['quantile_low']
            high = check_repair_config['miss_repair']['quantile_high']
            if not (0.0 <= low <= high <= 1.0):
                raise FeatureProcessException(
                    f"origin_col_num:{index}, quantile_low and quantile_high must between 0.0 and 1.0, and quantile_low must be smaller than quantile_high ")
            val = col_frame.quantile(
                low) + (col_frame.quantile(high) - col_frame.quantile(low)) * random.random()
            if check_repair_config['type'].lower() in ['int', 'integer']:
                val = int(val)
            col_frame = col_frame.fillna(val)
    return col_frame


def numeric_outlier_check(check_repair_config, col_frame, mode):
    index = check_repair_config['index']
    if 'outlier_check' not in check_repair_config:
        raise FeatureProcessException(
            f"origin_col_num:{index}, outlier_check config missing!")
    if not (check_repair_config['outlier_check'] in ['1', '2', '3']):
        raise FeatureProcessException(
            f"origin_col_num:{index}, outlier_check must be in ['1', '2', '3']!")

    low, high = 0, 0
    # 3 西格玛
    if check_repair_config['outlier_check'] == '1':
        mean = col_frame.mean()
        std = col_frame.std()
        low = mean - 3 * std
        high = mean + 3 * std
    elif check_repair_config['outlier_check'] == '2':
        low = col_frame.quantile(
            0.25) - 1.5 * (col_frame.quantile(0.75) - col_frame.quantile(0.25))
        high = col_frame.quantile(
            0.75) + 1.5 * (col_frame.quantile(0.75) - col_frame.quantile(0.25))
    return low, high


def numeric_outlier_repair(check_repair_config, col_frame, mode):
    index = check_repair_config['index']
    if not ('outlier_repair' in check_repair_config):
        raise FeatureProcessException(
            f"origin_col_num:{index}, outlier_repair config missing!")
    if not (check_repair_config['outlier_repair'] in ['1', '2', '3']):
        raise FeatureProcessException(
            f"origin_col_num:{index}, outlier_repair must in ['1', '2', '3']!")

    if check_repair_config['outlier_repair'] == '1':
        mean = col_frame.mean()
        std = col_frame.std()
        low = mean - 3 * std
        high = mean + 3 * std

        col_frame[col_frame < low] = low
        col_frame[col_frame > high] = high
    elif check_repair_config['outlier_repair'] == '2':
        low = col_frame.quantile(
            0.25) - 1.5 * (col_frame.quantile(0.75) - col_frame.quantile(0.25))
        high = col_frame.quantile(
            0.75) + 1.5 * (col_frame.quantile(0.75) - col_frame.quantile(0.25))

        col_frame[col_frame < low] = low
        col_frame[col_frame > high] = high
    return col_frame


def category_miss_check_repair(check_repair_config, col_frame, mode):
    group_cols = check_repair_config['group_cols']
    index = check_repair_config['index']
    data_frame = check_repair_config['data_frame']

    if check_repair_config['miss_repair'] == '4':
        col_frame = col_frame.fillna(col_frame.mode()[0])
    elif check_repair_config['miss_repair'] == '6':
        if not (group_cols is not None and type(group_cols) == list and len(group_cols) != 0):
            raise FeatureProcessException(
                f"origin_col_num:{index}, when miss_repair='6', group_cols can't be empty")
        groups = data_frame.groupby(group_cols, observed=True).groups
        for group, idxs in groups.items():
            df_tmp = col_frame.loc[idxs].copy()
            if len(df_tmp[~df_tmp.isnull()]) == 0:
                fill_item = col_frame.mode()[0]  # 如果类内全是null，则用全局替代
            else:
                fill_item = df_tmp[~df_tmp.isnull()].mode()[0]
            col_frame.loc[idxs] = col_frame.loc[idxs].fillna(fill_item)
    return col_frame


# todo: 日期类型特征当前没做处理，后期增补内容
class UnknownFeature(Feature):
    def __init__(self, index, config, mode, **kwargs):
        super(UnknownFeature, self).__init__(index, config, mode, **kwargs)
        self.fe_name2func = {}

    def do_process(self, data_frame):
        feat_df = data_frame[data_frame.columns[self.index]]

        cols = []
        if len(feat_df.shape) == 1:
            cols = [f'{self.index}_0']

        elif len(feat_df.shape) == 2:
            for i in range(feat_df.shape[1]):
                cols.append(f'{self.index}_{i}')

        feat_df.columns = cols
        # return feat_df, {}

        # Todo： 无法识别的类型，先drop掉
        return pd.DataFrame(), {}


def ordinal(params, col_frame, mode):
    if mode == 'train':
        category_list = col_frame.value_counts().index.tolist()
        params['category_list'] = category_list
    else:
        category_list = params['category_list']
    col_frame[~col_frame.isin(category_list)] = len(category_list)
    for i, cate in enumerate(category_list):
        col_frame[col_frame == cate] = i
    col_frame = col_frame.astype(int)
    return col_frame, params


def one_hot(params, col_frame, mode):
    if mode == 'train':
        category_list = col_frame.value_counts().index.tolist()
        params['category_list'] = category_list
    else:
        category_list = params['category_list']

    col_one_hot = np.zeros(
        (col_frame.shape[0], len(category_list) + 1), dtype=np.int32)

    for i, val in enumerate(col_frame.tolist()):
        if val == val and val in category_list:  # not nan
            col_one_hot[i][category_list.index(val)] = 1
        else:
            col_one_hot[i][len(category_list)] = 1

    frame_new = pd.DataFrame(col_one_hot, index=col_frame.index)
    return frame_new, params


def quantile_transform(params, col_frame, mode):
    if mode == 'train':
        bucket_size = params.get('bucket_size', 10)
        boundaries = []
        for pos in range(1, bucket_size):
            boundaries.append(
                col_frame[~col_frame.isnull()].quantile(pos / bucket_size))
        params['boundaries'] = boundaries
    else:
        boundaries = params['boundaries']

    bins = [float('-inf')]
    bins.extend(boundaries)
    bins.append(float('inf'))

    col_frame[~col_frame.isnull()] = pd.cut(col_frame[~col_frame.isnull()],
                                            bins, labels=False, duplicates='drop').astype(int)
    return col_frame, params


def log_transform(params, col_frame, mode):
    if not (col_frame[(~col_frame.isnull()) & (col_frame < 0)].empty):
        raise FeatureProcessException(
            'log_transform transform can only handle the data bigger than zero! while there are negetive data in dataframe')
    col_frame[~col_frame.isnull()] = np.log10(col_frame[~col_frame.isnull()] + 1)
    return col_frame, params


def standard_scale(params, col_frame, mode):
    if mode == 'train':
        mean = col_frame[~col_frame.isnull()].mean()
        std = col_frame[~col_frame.isnull()].std()
        params['mean'] = mean
        params['std'] = std
    else:
        mean = params['mean']
        std = params['std']
    if std == 0:
        col_frame.iloc[:, ] = 1
    else:
        col_frame[~col_frame.isnull()] = (
                                                 col_frame[~col_frame.isnull()] - mean) / std
    return col_frame, params


def max_abs_scale(params, col_frame, mode):
    if mode == 'train':
        max_abs = col_frame[~col_frame.isnull()].abs().max()
        params['max_abs'] = max_abs
    else:
        max_abs = params['max_abs']

    if max_abs == 0:
        col_frame.iloc[:, ] = 0
    else:
        col_frame[~col_frame.isnull()] = col_frame[~col_frame.isnull()] / max_abs
    return col_frame, params


def min_max_scale(params, col_frame, mode):
    if mode == 'train':
        max = col_frame[~col_frame.isnull()].max()
        min = col_frame[~col_frame.isnull()].min()
        params['max'] = max
        params['min'] = min
    else:
        max = params['max']
        min = params['min']
    if max == min:
        if max == 0:
            col_frame[~col_frame.isnull()].iloc[:, ] = 0
        else:
            col_frame[~col_frame.isnull()].iloc[:, ] = 1
    else:
        col_frame[~col_frame.isnull()] = (
                                                 col_frame[~col_frame.isnull()] - min) / (max - min)
    return col_frame, params


def bucket(params, col_frame, mode):
    if not ('boundaries' in params and len(params['boundaries']) > 0):
        raise FeatureProcessException(
            "for bucket transform ,boundaries can't be empty!")
    bins = [float('-inf')]
    bins.extend(params['boundaries'])
    bins.append(float('inf'))
    labels = range(len(bins) - 1)
    col_frame[~col_frame.isnull()] = pd.cut(col_frame[~col_frame.isnull()],
                                            bins, labels=False, duplicates='drop').astype(int)
    return col_frame, params


def translation(params, col_frame, mode):
    if mode == 'train':
        min = col_frame[~col_frame.isnull()].min()
        params['min'] = min
    else:
        min = params['min']
    if min >= 0:
        return col_frame, params
    col_frame[~col_frame.isnull()] = col_frame[~col_frame.isnull()] + abs(min)
    return col_frame, params


def square(params, col_frame, mode):
    col_frame[~col_frame.isnull()] = np.square(col_frame[~col_frame.isnull()])
    return col_frame, params


def robust_scale(params, col_frame, mode):
    if mode == 'train':
        median = col_frame[~col_frame.isnull()].quantile()
        quant25 = col_frame[~col_frame.isnull()].quantile(0.25)
        quant75 = col_frame[~col_frame.isnull()].quantile(0.75)
        params['median'] = median
        params['quant25'] = quant25
        params['quant75'] = quant75
    else:
        median = params['median']
        quant25 = params['quant25']
        quant75 = params['quant75']
    if quant25 != quant75:
        col_frame[~col_frame.isnull()] = (
                                                 col_frame[~col_frame.isnull()] - median) / (quant75 - quant25)
    return col_frame, params


def drop(params, col_frame, mode):
    return pd.DataFrame(), params


def hash_feature(params, col_frame, mode):
    if mode == 'train':
        params['bucket_size'] = params.get('bucket_size', 10)
    hash_bucket_size = params['bucket_size']
    col_frame[~col_frame.isnull()] = col_frame[~col_frame.isnull()].apply(
        lambda x: hash(x) % hash_bucket_size)
    return col_frame.astype(int), params


def interval_bucket(params, col_frame, mode):
    if mode == 'train':
        bucket_size = params.get('bucket_size', 10)
        boundaries = []
        maxim, minim = col_frame[~col_frame.isnull()].max(
        ), col_frame[~col_frame.isnull()].min()
        for pos in range(1, bucket_size):
            boundaries.append((maxim - minim) * pos / bucket_size + minim)
        params['boundaries'] = boundaries
    else:
        boundaries = params['boundaries']

    bins = [float('-inf')]
    bins.extend(boundaries)
    bins.append(float('inf'))

    col_frame[~col_frame.isnull()] = pd.cut(col_frame[~col_frame.isnull()],
                                            bins, labels=False, duplicates='drop').astype(int)
    return col_frame, params


def kmeans_bucket(params, col_frame, mode):
    data = col_frame[~col_frame.isnull()].to_numpy().reshape(-1, 1)
    if mode == 'train':
        n_clusters = params.get('bucket_size', 10)
        kmeans = KMeans(n_clusters=n_clusters).fit(data)
        centers = kmeans.cluster_centers_
        params['cluster_centers'] = centers
    else:
        centers = params['cluster_centers']

    dist = []
    for c in centers:
        dist.append(abs(data - c).reshape(-1))
    label = np.array(dist).argmin(axis=0)

    col_frame[~col_frame.isnull()] = label
    return col_frame, params



'''
Author: hexu
Date: 2021-10-14 14:54:05
LastEditTime: 2023-05-25 15:26:26
LastEditors: Hexu
Description: 数据集
FilePath: /iw-algo-fx/intelliw/datasets/datasets.py
'''
import json
from typing import List, Tuple,Union, overload
from intelliw.datasets.spliter import get_set_spliter
from intelliw.datasets.datasource_base import DatasetSelector, AbstractDataSource, AbstractDataSourceWriter, \
    DataSourceType as DST, DatasetType, AlgorithmsType
from intelliw.datasets.datasource_empty import EmptyDataSourceWriter
from intelliw.utils.logger import _get_framework_logger
from intelliw.config import config
from intelliw.utils.global_val import gl
from intelliw.datasets.dataset_config import DatasetConfigParams, convert
from dataclasses import asdict


logger = _get_framework_logger()


class DataSets:
    """
    DataSets
    dataset metadata
    """

    def __init__(self, datasource: AbstractDataSource, source_type: int, is_streaming=False):
        self.datasource = datasource
        self.alldata = []
        self.column_meta = []
        self.source_type = source_type
        self.model_type = gl.model_type  # 分类/回归/ocr/时间序列/文本分类。。。。。
        self.is_streaming = is_streaming

    def empty_reader(self, dataset_type=DatasetType.TRAIN):
        """
        empty reader
        """
        return self.datasource.reader(page_size=1, offset=0, limit=0, transform_function=None,
                                      dataset_type=dataset_type)

    def reader(self, page_size=10000, offset=0, limit=0, split_transform_function=None):
        """
        reader
        """
        return self.datasource.reader(page_size, offset, limit, split_transform_function)

    @overload
    def data_pipeline(self, split_transform_function,
                      alldata_transform_function, feature_process):
        """
        pipeline: download -> filter -> spliter -> user
        """
        pass

    def data_pipeline(self, *args):
        """
        pipeline: download -> filter -> spliter -> user
        """
        if self.source_type in DST.NLP_TYPE_LIST:
            return self._nlp_data(config.DATA_SPLIT_MODE)
        if self.source_type in DST.LARGE_MODEL_LIST:
            return self._large_model_data(config.DATA_SPLIT_MODE)
        elif self.source_type in DST.CV_TYPE_LIST:
            return self._images_data(*self._data_pipeline(*args))
        else:
            train, validation, test = self._data_pipeline(*args)
            return [train], [validation], [test]

    def read_all_data(self, split_transform_function=None):
        """
        download all data
        """
        reader = self.reader(config.DATA_SOURCE_READ_SIZE, 0,
                             self.datasource.total(), split_transform_function)
        for idx, content in enumerate(reader):
            if self.source_type not in DST.CV_TYPE_LIST:
                if idx == 0:
                    self.column_meta = reader.meta
                    self.alldata = content
                elif 'result' in content and 'result' in self.alldata:
                    self.alldata['result'].extend(content['result'])
            else:
                self.alldata.extend(content)
        return self.alldata

    def _data_pipeline(self, stf, atf, fp, ignore_dp=False, ignore_split=False):
        # 获取全部数据(切片数据处理， 列选择和数据筛选)
        alldata = self.read_all_data(stf)

        if not alldata:
            logger.warning("Data is empty")

        # 数据处理（时间序列，全局函数和特征工程）
        _data_process_args = [alldata, atf, fp]
        if ignore_dp:
            _data_process_args.append(True)
        alldata = self._data_process(*_data_process_args)

        # 数据集切分
        _get_set_spliter_args = [alldata, self.source_type]
        if ignore_split:
            _get_set_spliter_args.append(True)
        spliter = get_set_spliter(*_get_set_spliter_args)

        # 数据集处理 图片下载/语料下载/数据返回
        return spliter.train_reader(), spliter.validation_reader(), spliter.test_reader()

    def _data_process(self, alldata, atf, fp, do_nothing=False):
        """
        download -> filter -> spliter -> user
        """
        if atf is None and fp is None:
            do_nothing = True

        if do_nothing is True:
            pass
        elif self.source_type in DST.CV_TYPE_LIST:
            pass
        elif atf or fp:
            alldata = atf(alldata) if atf else alldata
            alldata = fp(alldata) if fp else alldata
        return alldata

    def _images_data(self, train, val, test):
        tr = self.datasource.download_images(
            train, dataset_type=DatasetType.TRAIN)
        v = self.datasource.download_images(
            val, dataset_type=DatasetType.VALID)
        te = self.datasource.download_images(
            test, dataset_type=DatasetType.TEST)
        return tr, v, te

    def _nlp_data(self, split_mode: int):
        self.datasource.corpora_process(split_mode)
        return [self.datasource()] * 3

    def _large_model_data(self, split_mode: int):
        self.datasource.corpora_process(split_mode)
        return self.datasource()

    def support_streaming(self):
        return self.source_type in DST.SUPPORT_STREAMING

class MultipleDataSets:
    def __init__(self) -> None:
        self._total = 0
        self.datasets: List[DataSets] = []
        self.join_type = "no"
        self.model_type = gl.model_type  # 分类/回归/ocr/时间序列/文本分类。。。。。
        self.column_meta = []
        self.is_streaming = False

    @property
    def total(self):
        return self._total

    @property
    def onlyone(self):
        return self._total == 1

    def add(self, dataset: DataSets):
        self.datasets.append(dataset)
        self._total += 1

    def pop(self, idx=None):
        if idx is not None and isinstance(idx, int):
            return self.datasets.pop(idx)
        return self.datasets.pop()

    @overload
    def data_pipeline(self, split_transform_function,
                      alldata_transform_function, feature_process):
        pass

    def read_all_data(self, split_transform_function=None):
        # TODO 多数据集暂时不支持特征工程
        return [d.read_all_data() for d in self.datasets]

    def data_pipeline(self, *args):
        result = None
        # TODO 这块的配置不支持混合模式
        for idx in range(self._total):
            dataset = self.pop(0)
            if dataset.source_type in DST.NLP_TYPE_LIST:
                result = dataset._nlp_data(
                    config.DATA_SPLIT_MODE
                )
            elif dataset.source_type in DST.LARGE_MODEL_LIST:
                result = dataset._large_model_data(
                    config.DATA_SPLIT_MODE
                )
            elif dataset.source_type in DST.CV_TYPE_LIST:
                result = dataset._images_data(
                    *dataset._data_pipeline(*args)
                )
            else:
                if not result:
                    result = [[None] * self._total for _ in range(3)]

                gl.dataset_idx = idx
                # TODO 多数据集暂时不支持特征工程
                this_args = list(args) + [True]
                result[0][idx], result[1][idx], result[2][idx] = dataset._data_pipeline(
                    *this_args)
                self.column_meta.extend(dataset.column_meta)
        return result


def get_dataset(cfg) -> Union[DataSets, MultipleDataSets]:
    """
    use dataset config to get dataset inforew

    """
    if not cfg:
        dataset_conf = [{"SOURCE_TYPE": DST.EMPTY}]
    elif isinstance(cfg, str):
        dataset_conf = json.loads(cfg)
    else:
        dataset_conf = cfg

    # 格式校验
    assert isinstance(dataset_conf, list), "dataset config must list"

    # table, cv, nlp
    mds = MultipleDataSets()
    for conf in dataset_conf:
        datasource, stype, is_streaming = DatasetSelector.parse(conf)
        mds.add(DataSets(datasource, stype, is_streaming))
    if mds.onlyone:
        return mds.pop()
    return mds

def get_dataset_by_params(params: DatasetConfigParams) -> Union[DataSets, MultipleDataSets]:
    """
    use dataset config to get dataset info
    """
    data_cfg = [asdict(convert(params))]
    logger.info("dataset config: {}".format(data_cfg))
    return get_dataset(data_cfg)

def get_datasource_writer(cfg: str = None) -> AbstractDataSourceWriter:
    """
    use dataset config to get dataset writer info
    """
    output_datasource_type = 0
    cfg = cfg or config.OUTPUT_DATASET_INFO
    if cfg:
        if isinstance(cfg, str):
            cfg = json.loads(cfg)
        output_datasource_type = cfg['sourceType']
    if output_datasource_type == DST.EMPTY:
        return EmptyDataSourceWriter()
    elif output_datasource_type in (DST.INTELLIV, DST.IW_FACTORY_DATA, DST.IW_FACTORY_DR_DATA):
        from intelliw.datasets.dataset_writer import DataSourceWriter
        return DataSourceWriter(output_config=cfg, writer_type=output_datasource_type)
    else:
        err_msg = f"输出数据源设置失败，无效的数据源类型: {output_datasource_type}"
        raise ValueError(err_msg)



import asyncio
import time
import types
import threading
import starlette.datastructures
from fastapi import Request as FARequest, Response as FAResponse
from fastapi.responses import StreamingResponse

import json
import traceback
from starlette.middleware.base import BaseHTTPMiddleware

from intelliw.utils import health_check
from intelliw.utils import message, aiocache
from intelliw.utils.aiocache import serializers
from intelliw.utils.intelliwapi import request as my_request, monitor
from intelliw.utils.context import header_ctx
from intelliw.utils.exception import LimitConcurrencyError
from intelliw.utils.logger import _get_framework_logger, inject_level, LogCfg

logger = _get_framework_logger()


class BaseHandler:
    """
    BaseHandler
    """

    def __init__(self, request: FARequest):
        self.infer_request = my_request.Request()
        self.fa_request = self.infer_request.raw = request

    async def handle_form_data(self):
        form_data = await self.fa_request._get_form()
        self.infer_request.form.get_dict = getattr(form_data, "_dict")
        for k, v in getattr(form_data, "_list"):
            if isinstance(v, starlette.datastructures.UploadFile):
                self.infer_request.files[k] = v
            _data = self.infer_request.form.get(k, [])
            _data.append(v)
            self.infer_request.form[k] = _data

    async def handle_json_data(self):
        self.infer_request.json = await self.fa_request.json()

    async def try_handle_json_data(self):
        try:
            self.infer_request.json = await self.fa_request.json()
        except json.decoder.JSONDecodeError:
            logger.error(f"request body JSONDecodeError, {traceback.format_exc()}")

    async def request_process(self):
        """
        Process incoming request data.

        Returns:
            Tuple[Union[Dict[str, Any], APIResponse], bool]: Parsed request data and success flag.
        """
        is_ok = True
        req_data = {}

        try:
            self.infer_request.header = self.fa_request.headers
            self.infer_request.method = self.fa_request.method
            self.infer_request.url = self.fa_request.url

            # Query parameters
            self.infer_request.query = self.fa_request.query_params._dict

            # Request body
            self.infer_request.body = await self.fa_request.body()
            content_type = self.infer_request.header.get('Content-Type', "").strip()

            # form
            if content_type.startswith('application/x-www-form-urlencoded') \
                    or content_type.startswith('multipart/form-data'):
                await self.handle_form_data()
                req_data = self.infer_request.form
            # json
            elif content_type.startswith('application/json') and self.infer_request.body:
                await self.handle_json_data()
                req_data = self.infer_request.json
            # body
            elif self.infer_request.body:
                await self.try_handle_json_data()
                req_data = self.infer_request.json

            # If the body is empty, try to get data from query parameters
            if not req_data:
                req_data = self.infer_request.query
        except Exception as e:
            logger.error(traceback.format_exc())
            msg = f"request解析错误: {e}, Body: {str(self.infer_request.body)}"
            req_data = message.APIResponse(400, "api", msg, msg)
            is_ok = False

        return req_data, is_ok

    async def response_process(self, data, func, need_feature):
        """
        response process
       """
        content_type = self.infer_request.header.get('Content-Type', "").strip()
        accept_header = self.infer_request.header.get('Accept', "").strip()
        request_path = str(self.infer_request.url.path) if hasattr(self.infer_request, 'url') and self.infer_request.url else ""
        request_method = str(self.infer_request.method).upper() if hasattr(self.infer_request, 'method') else ""
        
        # 修复流式请求判断逻辑：
        # 1. 请求的Content-Type包含 text/event-stream（原有逻辑保持兼容）
        # 2. 请求的Accept头包含 text/event-stream（SSE标准做法）
        # 3. 请求数据包含 stream 参数
        
        # 基础流式判断条件
        is_stream = (
            content_type.startswith("text/event-stream") or 
            accept_header.startswith("text/event-stream") or
            "text/event-stream" in accept_header )

        try:
            result, emsg = await self.infer(data, func, need_feature, )
            if emsg is None:
                # 只有满足流式条件且返回生成器时才返回流式响应
                if is_stream and type(result) in (types.GeneratorType, types.AsyncGeneratorType):
                    return result, True
                
                # MCP请求特殊处理：非流式MCP请求返回原始JSON-RPC格式
                if ("/mcp/" in request_path or request_path.endswith("/mcp")) and not is_stream:
                    # 对于MCP请求，如果result是字典且包含JSON-RPC字段，直接返回原始格式
                    if isinstance(result, dict) and ('id' in result or 'method' in result or 'result' in result or 'error' in result):
                        return result, False
                
                resp = message.APIResponse(200, "api", '', result)
            elif isinstance(emsg, LimitConcurrencyError):
                resp = message.APIResponse(message.ErrorCode.limit_exceeded, "api", result)
            else:
                resp = message.APIResponse(message.ErrorCode.invalid_request, "api", emsg, result)
        except Exception as e:
            logger.error(traceback.format_exc())
            msg = f"API服务处理推理数据错误 {e}"
            resp = message.APIResponse(message.ErrorCode.invalid_request, "api", msg, msg)
        return resp, False


    @aiocache.cached(key_builder=aiocache.common.key_builder,
                     skip_cache_func=aiocache.common.skip_cache_func,
                     cache=aiocache.Cache.REDIS, namespace="intelliw:N:cache",
                     serializer=serializers.PickleSerializer())
    async def infer(self, data, func, need_feature):
        return await self.fa_request.app.intelliw_setting["infer"].infer(
            data, self.infer_request, func, need_feature, )


class RequestTimingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: FARequest, call_next):
        request.state.req_start_time = time.time()
        request.state.ingress_start_time = request.headers.get('x-ingress-start-time')
        keywords = {'traceId': request.headers.get('traceId', '0')}

        yat = request.headers.get('yht_access_token')
        cookie = request.headers.get('cookie')
        ContentType = request.headers.get('Content-Type') 
        Accept = request.headers.get('Accept')
        sessionId = request.headers.get('sessionId')

        if yat:
            keywords['yht_access_token'] = yat
        if cookie:
            keywords['cookie'] = cookie
        # 额外透传一些头信息
        if ContentType:
            keywords['ContentType'] = ContentType
        if Accept:
            keywords['Accept'] = Accept
        if sessionId:
            keywords['sessionId'] = Accept
        
        logger.debug(f'request headers keywords: {keywords}')
        header_ctx.set(keywords)

        response = await call_next(request)
        response.headers.update(keywords)
        return response


def get_api_config(request: FARequest):
    api_config = request.app.intelliw_setting[f'api-{request.url.path}']
    return api_config['func'], api_config.get('need_feature', False)


async def api_handler(request: FARequest):
    base = BaseHandler(request)
    request_data, is_ok = await base.request_process()
    result = {}
    # 进行推理处理
    if is_ok:
        func, need_feature = get_api_config(request)
        result, is_stream = await base.response_process(request_data, func, need_feature)
        if is_stream:
            return StreamingResponse(result, media_type="text/event-stream")
    
    # 检查是否为MCP请求返回的原始JSON-RPC格式
    request_path = str(request.url.path) if hasattr(request, 'url') and request.url else ""
    if ("/mcp/" in request_path or request_path.endswith("/mcp")):
        # 对于MCP请求，如果result是字典且包含JSON-RPC字段，直接序列化为JSON
        if isinstance(result, dict) and ('id' in result or 'method' in result or 'result' in result or 'error' in result):
            return FAResponse(content=json.dumps(result), media_type="application/json")
        # 如果result是APIResponse对象，提取其原始数据
        elif hasattr(result, '_format'):
            formatted_result = result._format()
            # 检查是否包含extrainfo和data字段，如果是则提取data字段
            if isinstance(formatted_result, dict) and 'extrainfo' in formatted_result and 'data' in formatted_result:
                data_content = formatted_result['data']
                # 如果data是JSON-RPC格式，直接返回
                if isinstance(data_content, dict) and ('id' in data_content or 'method' in data_content or 'result' in data_content or 'error' in data_content):
                    return FAResponse(content=json.dumps(data_content), media_type="application/json")
            return FAResponse(content=json.dumps(formatted_result), media_type="application/json")
    
    if request.query_params.get('request_debug'):
        logger.warning(f'Request: {request_data}\nResponse: {result}')
    return FAResponse(content=str(result), media_type="application/json")


async def health_check_handler():
    loop = asyncio.get_event_loop()
    result = await loop.run_in_executor(None, health_check.health_check_process)
    if result:
        resp = message.HealthCheckResponse(200, "api", 'ok', "")
    else:
        resp = message.HealthCheckResponse(500, "api", 'error', "")
    return FAResponse(content=str(resp), media_type="application/json")


async def set_config(request: FARequest):
    try:
        data = await request.json()
        key = data.get('key')
        if key == 'log':
            level, delay = data.get('level', 'INFO'), min(data.get('delay', 600), 3600)
            inject_level(level)
            threading.Timer(delay, inject_level, (LogCfg.level,)).start()
        resp = {'code': 200, 'message': 'ok'}
    except Exception as e:
        logger.error(traceback.format_exc())
        resp = {'code': message.ErrorCode.invalid_request, 'message': str(e)}
    return FAResponse(content=str(resp), media_type="application/json")


async def service_monitoring(request: FARequest):
    data = await request.json()
    response_data = monitor.init_request_data(data)
    if response_data:
        return FAResponse(content=str(message.APIResponse(200, "api", '', response_data)),
                          media_type="application/json")
    # 查询
    if data.get('requestMode') == "query":
        response_data = monitor.monitor_query()
    # 请求监控
    elif data.get('requestMode') == "monitor":
        # 默认状态为监控任务运行中
        response_data = {"status": 1}
        monitor.monitor_request(data)

    resp = message.APIResponse(200, "api", '', response_data)
    return FAResponse(content=str(resp), media_type="application/json")


async def eureka_health_check_handler():
    resp = message.HealthCheckResponse(200, "api", 'ok', "")
    return FAResponse(content=str(resp), media_type="application/json")



'''
Author: hexu
Date: 2021-10-25 15:20:34
LastEditTime: 2023-05-24 15:15:48
LastEditors: Hexu
Description: api处理函数
FilePath: /iw-algo-fx/intelliw/interface/apihandler.py
'''
import os
import platform
from concurrent.futures import ThreadPoolExecutor

from fastapi import FastAPI

import json
import threading

from intelliw.config import config
from intelliw.core.infer import Infer
from intelliw.utils import message, util
from intelliw.interface.iwapi import iwhandler
from intelliw.utils import get_json_encoder
from intelliw.utils.intelliwapi import gunicorn_server
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


def set_initializer(func, *args):
    ApiService.initializer = _INITIALIZER(func, *args)


class _INITIALIZER:
    def __init__(self, func, *args):
        self.func = func
        self.args = args

    def run(self):
        return self.func(*self.args)


class Application:
    """推理服务路由类
    example:
        @Application.route("/infer-api", method='get', need_feature=True)
        def infer(self, test_data):
            pass
    args:
        path           访问路由   /infer-api
        method         访问方式，支持 get post push delete head patch options
        need_feature   是否需要使用特征工程, 如果是自定义与推理无关的函数, 请设置False
    """

    # Set URL handlers
    HANDLERS = []

    def __init__(self, custom_router):
        self.app = FastAPI(json_encoder=get_json_encoder())
        self.app.intelliw_setting = {}
        self.handler_process(custom_router)

    def __call__(self):
        return self.app

    @classmethod
    def route(cls, path, **options):
        """
        register api route
        """

        def decorator(function):
            cls.HANDLERS.append((
                path,
                {'func': function.__name__,
                 'method': options.pop('method', 'post').lower(),
                 'need_feature': options.pop('need_feature', True)}))
            return function

        return decorator

    def handler_process(self, routers):
        # 加载自定义api, 配置在algorithm.yaml中
        for router in routers:
            Application.HANDLERS.append((
                router["path"],
                {'func': router["func"],
                 'method': router.get("method", "post").lower(),
                 'need_feature': router.get("need_feature", True)}))

        # 检查用户是否完全没有配置路由
        if len(Application.HANDLERS) == 0:
            Application.HANDLERS.append((
                '/predict',
                {'func': 'infer', 'method': 'post', 'need_feature': True}))  # 默认值

        # 任务检测
        Application.HANDLERS.append((
            '/intelliw-worker',
            {'func': 'intelliw-worker', 'method': 'get', 'need_feature': False}
        ))

        # 集中绑定路由
        _route_cache = {}
        # 路由重复统计
        _route_dup_dict = dict()

        for router, info in Application.HANDLERS:
            # 确保路由以斜杠开头
            if not router.startswith("/"):
                router = "/" + router

            # 获取函数、方法及是否需要特征处理的信息
            func = info['func']
            methods = info['method'].split(",")
            need_feature = info['need_feature']

            # 检查路由是否已存在
            route_key = router + func
            if _route_cache.get(route_key, None):
                continue
            _route_cache[route_key] = True

            # 添加路由配置
            self.app.intelliw_setting[f'api-{router}'] = info
            self.app.add_api_route(router, endpoint=iwhandler.api_handler, methods=methods)

            # 记录并警告重复路由
            if router in _route_dup_dict:
                prev_func = _route_dup_dict[router]
                logger.warning("方法: %s, 访问路径：%s 将被 方法: %s 覆盖", prev_func, router, func)

            logger.info("方法: %s 加载成功, 访问路径：%s, 访问方法: %s, 是否需要特征处理: %s", func, router, methods,
                        need_feature)

        # healthcheck
        # gateway
        self.app.add_api_route(
            '/healthcheck',
            endpoint=iwhandler.health_check_handler, methods=['get', 'post'])

        # 专属化需求，用于检查python是否ready
        self.app.add_api_route(
            '/_python/ready',
            endpoint=iwhandler.health_check_handler, methods=['get', 'post'])
        # eureka
        self.app.add_api_route(
            '/CloudRemoteCall/',
            endpoint=iwhandler.eureka_health_check_handler, methods=['get', 'post'])
        # serviceMonitoring
        self.app.add_api_route(
            '/serviceMonitoring',
            endpoint=iwhandler.service_monitoring, methods=['post'])
        # set config
        self.app.add_api_route(
            '/config',
            endpoint=iwhandler.set_config, methods=['post'])

        # middleware
        self.app.add_middleware(iwhandler.RequestTimingMiddleware)


class ApiService:
    """
    intelliw api service
    """
    _instance = None
    _init_flag = False

    initializer = None

    def __new__(cls, *args, **kwargs):
        # 1.判断类属性是否为空对象，若为空说明第一个对象还没被创建
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self, port, path, response_addr):
        if ApiService._init_flag:
            return
        ApiService._init_flag = True
        self.port = port  # 8888
        self.infer = Infer(path, response_addr)
        self.app = Application(self.infer.pipeline.custom_router).app
        self.reporter = self.infer.pipeline.recorder
        self.app.intelliw_setting.update({"infer": self.infer, "reporter": self.reporter})
        self._report_start()

    def _report_start(self):
        """
        report start
        """
        self.reporter.report(message.CommonResponse(
            200, "inferstatus", '',
            json.dumps([{'status': 'start',
                         'inferid': config.INFER_ID,
                         'instanceid': config.INSTANCE_ID,
                         'inferTaskStatus': []}],
                       cls=get_json_encoder(), ensure_ascii=False)
        ))

    @staticmethod
    def _eureka_server():
        if len(config.EUREKA_SERVER) > 0:
            from intelliw.core.linkserver import linkserver
            try:
                should_register = config.EUREKA_APP_NAME != ''
                iports = json.loads(config.REGISTER_CLUSTER_ADDRESS)
                profile = config.EUREKA_ZONE or 'test'
                linkserver.client(
                    config.EUREKA_SERVER, config.EUREKA_PROVIDER_ID,
                    should_register, config.EUREKA_APP_NAME, iports, profile)
                logger.info("eureka server client init success, register:%s, server name: %s",
                            should_register, config.EUREKA_APP_NAME)
            except Exception as e:
                logger.error(
                    f"eureka server client init failed, error massage: {e}")

    def thread_pool_init(self, thread):
        # 默认放开线程限制
        recommended_thread = thread
        
        # 优先使用用户定义的线程数
        thread_define = os.environ.get('THREAD_USE_DEFINE')
        if thread_define:
            try:
                thread = int(thread_define)
                logger.info(f"Using user-defined thread count: {thread}, recommended: {recommended_thread}")
            except ValueError:
                logger.error(f"Invalid THREAD_USE_DEFINE value: {thread_define}, using unlimited thread count instead")
                thread = 999
                logger.info(f"Using unlimited thread count: {thread}, recommended: {recommended_thread}")
        elif os.environ.get('THREAD_USE_RECOM') == 'true':
            logger.info(f"Using recommended thread count: {thread}")
        else:
            thread = 999
            logger.info(f"Using unlimited thread count: {thread}, recommended: {recommended_thread}")
            logger.info("If you encounter issues, you can try:")
            logger.info("1. Set THREAD_USE_RECOM=true to use the recommended thread count")
            logger.info("2. Set THREAD_USE_DEFINE=<number> to specify a custom thread count")

        initializer, initargs = None, ()
        if self.initializer is not None:
            def threading_init(pool, func, *args):
                t = threading.currentThread().ident
                pool[t] = func(*args)

            initializer = threading_init
            initargs = (
                self.infer.pipeline.thread_data_pool,
                self.initializer.func,
                *self.initializer.args
            )

        self.infer.pipeline.max_wait_task = min(999, max(2, config.INFER_MAX_TASK_RATIO * thread))
        self.infer.pipeline.async_executor = ThreadPoolExecutor(
            thread, initializer=initializer, initargs=initargs, thread_name_prefix="IntelliwServerPool")
        return thread

    def _fastapi_server(self):
        config.MAINPID = str(os.getpid())

        worker, thread = util.get_worker_count(
            config.CPU_COUNT,
            config.INFER_MULTI_THREAD_COUNT,
            config.INFER_MULTI_PROCESS
        )

        config.INFER_MULTI_PROCESS_COUNT = worker
        thread = self.thread_pool_init(thread)

        # isIpv6 = util.use_ipv6_by_env()
        # 多进程
        if config.INFER_MULTI_PROCESS:
            setting = gunicorn_server.default_config(
                # f'[::]:{self.port}' if isIpv6 else f'0.0.0.0:{self.port}',
                f'[::]:{self.port}',
                workers=worker,
            )
            server = gunicorn_server.GunServer(self.app, setting, logger)
        else:
            # 默认模式
            server = gunicorn_server.UvicornServer(
                self.app,
                ["0.0.0.0", "::"] if platform.system() == "Windows" else "*",
                self.port
            )

        logger.info("\033[34mServer init success, server type: %s, workers: %s, threads: %s, max wait task: %s \033[0m",
                    server.__class__.__name__, worker, thread, self.infer.pipeline.max_wait_task)
        server.run()

    def run(self):
        """
        start server
        """
        self._eureka_server()
        self._fastapi_server()



"""
iUAP HTTP auth protocol implementation
"""
import json
import os
import time
import math
import json as _py_json
import hashlib
import base64
import logging
import uuid
import sys
import requests
import requests.exceptions
import requests.packages

from intelliw.config import config
from intelliw.utils.context import header_ctx
from intelliw.utils.trace_log import TraceLog

requests.packages.urllib3.disable_warnings()
logger = logging.getLogger("request")

try:
    import jwt

    has_jwt_package = True
except ImportError:
    logger.warning(
        "\033[33mIf want use authsdk, you need: pip install pyjwt\033[0m")
    has_jwt_package = False

DEFAULT_TIMEOUT: float = 10.0


def replace_url(url):
    internal_aliyun = "intelliw-console.oss-cn-beijing-internal.aliyuncs.com"
    yy_oss = "intelliw-console.oss-cn-beijing.aliyuncs.com"
    return url.replace(internal_aliyun, yy_oss)


class AuthType:
    """
    AuthType
    """
    No = 0
    AuthSDK = 1
    YHT = 2

    AuthCache = {}


class Response:
    """
    Response
    """

    def __init__(self, resp: requests.Response, error: Exception = None):
        self.raw = resp
        self.error = error
        self.text = self.raw.text
        self.status = self.status_code = self.raw.status_code
        self.content = self.body = self.raw.content
        self.json = self._try_json(self.raw.json)

    def raise_for_status(self):
        """
        raise :class:`IuapRequestException <IuapRequestException>` if status is not 200
        """
        if self.status != 200:
            msg = f'http request error, status: [{self.status}], body: [{self.body}]'
            if self.error is not None:
                raise IuapRequestException(
                    msg + str(self.error)) from self.error
            raise IuapRequestException(msg)

    @staticmethod
    def _try_json(f):
        try:
            return f()
        except Exception:
            return None

    def __str__(self):
        return f'status: {self.status}, body: {self.body}, error: {self.error}'


class IuapRequestException(Exception):
    """
    IuapRequestException
    """
    pass

def async_report_trace_log(tenant_id: str, trace_log: TraceLog):
    executor = trace_log._get_executor()
    executor.submit(report_trace_log, tenant_id, trace_log)

def report_trace_log(tenant_id: str, trace_log: TraceLog):
    """
    :param trace_log
    """

    try:
        # 加签
        headers = {"Content-Type": "application/json"}
        headers['X-tenantId'] = tenant_id
        auth_type = AuthType.AuthSDK
        url = f"{os.environ['domain.url']}/iuap-aip-console/self/rest/api/traceLogs"

        if trace_log.endTime is None:
            trace_log.endTime = int(time.time() * 1000)

        if trace_log.timeCost is None:
            trace_log.calculate_time_cost()

        data = trace_log.to_dict()


        if has_jwt_package:
            headers, params = sign(url, {}, headers, auth_type)
        # logger.info(f'trace log data:{data}')
        resp = __do_request(method="POST", url=url, headers=headers, params=params,
                            json=data, auth_type=auth_type, retry=2)
        logger.info("trace log, traceid: %s, resp: %s", trace_log.traceId, resp)
    except Exception as e:
        logger.error(f"report trace log error: {str(e)}")

def get_module_config(url, params = None):
    """
    :param url
    """
    # 加签
    headers = {
        "Content-Type": "application/json",
        "X-tenantId": '0'
        }
    auth_type = AuthType.AuthSDK

    if params is None:
        params = {}

    if has_jwt_package:
        headers, params = sign(url, params, headers, auth_type)
    resp = None
    try :
        resp = __do_request(method="GET", url=url, headers=headers, params=params, auth_type=auth_type)
        resp.raise_for_status()
        response_body = resp.json
        return resp.status, response_body
    except Exception as e:
        logger.warning(f"get module config error: {str(e)}")
        return resp.status if resp is not None and hasattr(resp, 'status') else None, e

def download(url, output_path=None, method="GET", params=None, data=None, json=None, headers=None,
             auth_type=AuthType.No, ak=None, ase=None, timeout=DEFAULT_TIMEOUT):
    """
    download file
    """
    # 加签
    if headers is None:
        headers = {}

    if ak is not None and ase is not None:
        auth_type = AuthType.AuthSDK

    if has_jwt_package and auth_type != AuthType.No:
        headers, params = sign(url, params, headers, auth_type, ak=ak, ase=ase)

    resp = __do_request(method=method, url=url, headers=headers,
                        data=data, params=params, json=json, auth_type=auth_type, timeout=timeout)
    if output_path is None:
        return resp
    else:
        resp.raise_for_status()
        if '<Error>' in resp.text:
            raise IuapRequestException(resp.text)
        mode = "w" if isinstance(resp.body, str) else "wb"
        with open(output_path, mode) as code:
            code.write(resp.body)

def get_modelfiles_by_modelid(url, model_id, timeout=10):
    """
    :param url
    """
    # 加签
    headers = {
        "Content-Type": "application/json",
        # "X-tenantId": '0'
        }
    auth_type = AuthType.AuthSDK
    params = {
        "modelId": model_id
    }

    if has_jwt_package:
        headers, params = sign(url, params, headers, auth_type)

    resp = __do_request(method="GET", url=url, headers=headers, params=params, auth_type=auth_type, timeout=timeout)
    resp.raise_for_status()
    response_body = resp.json
    return response_body    

def stream_download(url, output_path, method="get", params=None, body=None, json=None, headers=None,
                    auth_type=AuthType.No, ak=None, ase=None):
    """
    stream_download 流式下载文件
    """
    chunk_size = 1024

    def report_hook(plan, speed):
        logger.info("dataset downloading: {:.3f}% {:.3f}MB/s".format(
            plan, speed / (chunk_size ** 2)))

    # 加签
    if headers is None:
        headers = {}

    if ak is not None and ase is not None:
        auth_type = AuthType.AuthSDK

    if has_jwt_package and auth_type != AuthType.No:
        headers, params = sign(url, params, headers, auth_type, ak=ak, ase=ase)

    # 请求下载地址，以流式的。打开要下载的文件位置。
    with requests.request(method=method, url=url, stream=True, verify=False, data=body, params=params, json=json,
                          timeout=DEFAULT_TIMEOUT) as r, open(output_path, 'wb') as file:
        total_size = int(r.headers['content-length'])
        start_time = time.time()  # 请求开始的时间
        download_content_size = 0  # 下载的字节大小
        temp_size = 0  # 上秒的下载大小
        plan = 0  # 进度下载完成的百分比

        # 开始下载每次请求chunk_size字节
        for idx, content in enumerate(r.iter_content(chunk_size=chunk_size)):
            file.write(content)
            download_content_size += len(content)
            plan = (download_content_size / total_size) * 100
            if time.time() - start_time > 1:
                start_time = time.time()
                speed = download_content_size - temp_size
                if idx % 5 == 0:
                    report_hook(plan, speed)
                temp_size = download_content_size


def get(url: str, headers: dict = None, params: dict = None, timeout: float = DEFAULT_TIMEOUT,
        auth_type=AuthType.AuthSDK, ak=None, ase=None, retry=True) -> Response:
    """
    get request

    :param auth_type: 是否需要鉴权, 0 不需要, 1 authsdk, 2 yht
    :param timeout: request timeout
    :param url: request url
    :param headers: request headers
    :param params: request url params
    :param ak: 用户自定义access key
    :param ase: 用户自定义access secret
    :param retry: 重试, 0或false为不重试，可以设置重试次数
    :return: Response
    """
    return sign_and_request(url, 'GET', headers, params, timeout=timeout, auth_type=auth_type, ak=ak, ase=ase,
                            retry=retry)


def post_json(url: str, headers: dict = None, params: dict = None, json: object = None,
              timeout: float = DEFAULT_TIMEOUT, auth_type=AuthType.AuthSDK, ak=None, ase=None,
              retry: [bool, int] = True) -> Response:
    """
    post request, send data as json


    :param auth_type: 是否需要鉴权, 0 不需要, 1 authsdk, 2 yht
    :param timeout: request timeout
    :param url: request url
    :param headers: request headers
    :param params: request url parameters
    :param json: request body. if data is not `str`, it will be serialized as json.
    :param ak: 用户自定义access key
    :param ase: 用户自定义access secret
    :param retry: 重试, 0或false为不重试，可以设置重试次数
    :return: Response
    """

    if headers is None:
        headers = {}
    headers['Content-type'] = 'application/json; charset=UTF-8'
    return sign_and_request(url, 'POST', headers, params, json=json,
                            timeout=timeout, auth_type=auth_type, ak=ak, ase=ase, retry=retry)


def post(url: str, headers: dict = None, params: dict = None, data: object = None, json=None, files = None, 
         timeout: float = DEFAULT_TIMEOUT, auth_type=AuthType.AuthSDK, ak=None, ase=None,
         retry: [bool, int] = True) -> Response:
    """
    post request, send data


    :param auth_type: 是否需要鉴权 0 不需要 1 authsdk 2 yht
    :param timeout: request timeout
    :param url: request url
    :param headers: request headers
    :param params: request url parameters
    :param data: request body. if data is not `str`, it will be serialized as json.
    :param json: Request body in Json format
    :param files: Request body in file format
    :param ak: 用户自定义access key
    :param ase: 用户自定义access secret
    :param retry: 重试, 0或false为不重试，可以设置重试次数
    :return: Response
    """

    if headers is None:
        headers = {}

    if json is not None:
        headers['Content-type'] = 'application/json; charset=UTF-8'

    return sign_and_request(url, 'POST', headers, params, data=data, json=json, files=files,
                            timeout=timeout, auth_type=auth_type, ak=ak, ase=ase, retry=retry)


def put_file(url: str, headers: dict = None, params: dict = None, data: object = None, files = None,
             timeout: float = DEFAULT_TIMEOUT, auth_type=AuthType.No, ak=None, ase=None,
             retry: [bool, int] = True) -> Response:
    """
    put file
    """
    if headers is None:
        headers = {'Content-Type': 'application/octet-stream'}
    return sign_and_request(url, 'PUT', headers, params, data=data,files=files,
                            timeout=timeout, auth_type=auth_type, ak=ak, ase=ase, retry=retry)


def sign_and_request(url: str,
                     method: str = 'GET',
                     headers: dict = None,
                     params: dict = None,
                     data: bytes = None,
                     json: dict = None,
                     files: dict = None,
                     timeout: float = DEFAULT_TIMEOUT,
                     auth_type=AuthType.AuthSDK,
                     ak=None, ase=None, retry=True) -> Response:
    """
    sign and do request

    :param url: request url, without query
    :param method: Http request method, GET, POST...
    :param headers: request headers
    :param params: parameters will be sent as url parameters. Also used to generate signature if sign_params is None.
    :param data: request body
    :param json: Request body in Json format
    :param files: Request body in file format
    :param timeout: url access timeout
    :param auth_type: 是否需要鉴权 0 不需要 1 authsdk 2 yht
    :param ak: 用户自定义access key
    :param ase: 用户自定义access secret
    :param retry: 重试, 0或false为不重试，可以设置重试次数
    :return: response body
    """
    if headers is None:
        headers = {}

    # auth
    sign(url, params, headers, auth_type, ak=ak, ase=ase)
    return __do_request(url, method, headers, params, data, json, files = files, timeout=timeout, auth_type=auth_type, retry=retry)


def sign(url: str, params: dict, headers: dict, auth_type=AuthType.AuthSDK, ak=None, ase=None, refresh=False):
    if ak is not None and ase is not None:
        auth_type = AuthType.AuthSDK

    if has_jwt_package and auth_type == AuthType.AuthSDK:
        if params is None:
            params = {}
        params['AuthSdkServer'] = "true"
        headers['YYCtoken'] = sign_authsdk(url, params, ak=ak, ase=ase)
    elif auth_type == AuthType.YHT:
        yht_access_token = headers.get('yht_access_token')
        if yht_access_token is None or yht_access_token == '':
            yht_access_token = sign_yht(headers, refresh)
        # headers['cookie'] = f'yht_access_token={yht_access_token}'
        headers['yht_access_token'] = yht_access_token
        headers['cookie'] = None
    return headers, params


def sign_authsdk(url: str, params: dict, ak=None, ase=None) -> str:
    """
    generate iuap signature

    :param url: request url, without parameters
    :param params:  request parameters, x-www-form-urlencoded request's body parameters should also be included.
    :param ak: ACCESS KEY
    :param ase: ACCESS SECRET
    :return: iuap signature
    """
    if not ak:
        ak = os.environ.get('ACCESS_KEY') or config.ACCESS_KEY
    if not ase:
        ase = os.environ.get('ACCESS_SECRET') or config.ACCESS_SECRET

    issue_at = __issue_at()
    sign_key = __build_sign_key(ak, ase, issue_at, url)
    jwt_payload = {
        "sub": url,
        "iss": ak,
        "iat": issue_at
    }
    if params is not None and len(params) > 0:
        sorted_params = sorted(params.items())
        for item in sorted_params:
            if item[1] is None:
                val = ''
            elif len(str(item[1])) >= 1024:
                val = str(__java_string_hashcode(str(item[1])))
            else:
                val = str(item[1])
            jwt_payload[item[0]] = val

    jwt_token = jwt.encode(jwt_payload, key=sign_key, algorithm='HS256')
    return jwt_token if isinstance(jwt_token, str) else jwt_token.decode('utf-8')


def sign_yht(headers=None, refresh=False):
    # 本地获取会被运维拦住，只能手动复制一个
    if not config.is_server_mode():
        if config.TEMPORARY_USER_COOKIE:
            return config.TEMPORARY_USER_COOKIE
        else:
            raise SignError("为配置用户Token",
                            "本地调试数据集，请先在Cookie中获取用户Token, 并配置到环境变量 TEMPORARY_USER_COOKIE ")

    # 在线模式
    xTenantId = None
    TENANT_ID = os.environ.get('TENANT_ID') or config.TENANT_ID
    if headers is not None:
        xTenantId = headers.get('X-tenantId')
    
    if xTenantId is None:
        logger.debug(f"X-tenantId not found, use default tenantId:{TENANT_ID}")
        xTenantId =  TENANT_ID
    cache_key = f'yhtToken_{xTenantId}'
    token = AuthType.AuthCache.get(cache_key, {"expire": 0})
    expire = token.get("expire")

    if not refresh and (expire != 0 and time.time() < expire):
        logger.debug("Get Cache Key: %s, Cache Token: %s", cache_key, token)
        return token['token']
    else:
        # 请求 AI工作坊获取
        logger.debug(f"Get YHTToken request, yht_url: {config.GENERATOR_YHT_URL}, headers: {headers}")
        resp = get(config.GENERATOR_YHT_URL, headers)
        resp.raise_for_status()
        try:
            token = resp.json['data']['yhtToken']
            logger.debug("Get X-tenantId: %s, New Token: %s", xTenantId, token)
            AuthType.AuthCache[cache_key] = {
                'expire': time.time() + 60, "token": token}
        except Exception as e:
            msg = f"Get YHTToken Failed: {e}, resp body: [{resp.body}], resp json: {resp.json}"
            logger.error(msg)
            raise Exception(msg)
        return token


def __issue_at():
    issue_at = int(time.time())
    issue_at = math.floor(issue_at / 600) * 600
    return issue_at


def __build_sign_key(access_key, access_secret, access_ts, url):
    str_key = access_key + access_secret + str(access_ts * 1000) + url
    sign_key_bytes = hashlib.sha256(str_key.encode('UTF-8')).digest()
    return base64.standard_b64encode(sign_key_bytes).decode('UTF-8')


def __java_string_hashcode(s: str):
    h = 0
    for c in s:
        h = (31 * h + ord(c)) & 0xFFFFFFFF
    return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000


def __do_request(url: str,
                 method: str = 'GET',
                 headers: dict = {},
                 params: dict = {},
                 data: bytes = None,
                 json: object = None,
                 files: dict = None,
                 timeout: float = DEFAULT_TIMEOUT,
                 auth_type=AuthType.No, retry: [bool, int] = True) -> Response:
    # header format
    if headers is None:
        headers = {}

    xTenantId = headers.get('X-tenantId')
    if xTenantId is None:
        TENANT_ID = os.environ.get('TENANT_ID') or config.TENANT_ID
        headers = {
            **headers,
            **{'X-tenantId': TENANT_ID,
            'tenant_id': TENANT_ID,
            'tenantId': TENANT_ID,
            'instanceID': TENANT_ID,
            }
        }
    else:
        headers = {
            **headers,
            **{
            'tenant_id': xTenantId,
            'tenantId': xTenantId,
            'instanceID': xTenantId,
            }
        }

    retry_count = 4
    if type(retry) == int and retry > 0:
        retry_count = retry

    for i in range(1, retry_count + 1):
        try:
            if i == 2:
                url = replace_url(url)
            logger.debug("request: %s %s %s %s %s %s", method, url, params, data, json, headers)
            resp = requests.request(
                method=method, url=url, params=params, data=data,
                json=json, files=files, headers=headers, verify=False, timeout=timeout
            )
            body = resp.text
            # 验签失败重试
            if "<title>登录</title>" in body:
                raise SignError("跳转登陆页, 请重新登录", f"Token验证未通过, 跳转登陆页")
            return Response(resp)
        except (requests.exceptions.RequestException, SignError) as e:
            if not retry or i == retry_count:
                raise e
            time.sleep(2)
            headers, _ = sign(url, params, headers, auth_type, refresh=True)
            try:
                body = e.response.text if hasattr(e.response, "text") else e.response
            except:
                body = ""
            logger.warning(
                "request retry time: %s, url: %s, body: %s, error: %s",
                i, url, body, e)


class SignError(Exception):
    def __init__(self, body="", msg=""):
        self.response = body
        self.msg = msg

    def __str__(self):
        return self.msg

    @staticmethod
    def ignore_stack():
        return True


if __name__ == '__main__':
    # post_json("http://www.google.com", timeout=1, retry=5)
    now = time.time() * 1000
    # trace_log("hexutest", now, {"test": 1}, {"test": 2})



import threading
import os
import time
import requests
from typing import Dict, Optional, Any
from collections import OrderedDict
from intelliw.utils import iuap_request
from intelliw.config import config
import logging

logger = logging.getLogger('user_router')
logger.setLevel("INFO")
logger_code = "intelliw.logger.level"

class UserRouter:

    TENANT_ID_MAPPING_URI = "/self/tenantapi/acquireToken" # 获取租户ID映射关系

    def __init__(
        self,
        timeout: float = 5.0,
        cache_size: int = 1000,
        cache_ttl: int = 60  # 1分钟
    ):
        self.mapping_url = f"{os.environ.get('domain.url')}/{os.getenv('YMS_CONSOLE_APP_CODE', 'iuap-aip-console')}{UserRouter.TENANT_ID_MAPPING_URI}"
        self.timeout = timeout
        self.tenantid_cache_lock = threading.Lock()  # 专门用于tenantid缓存的锁
        self.token_cache_lock = threading.Lock()  # 专门用于token缓存的锁
        
        # tenantid缓存：{tenantid: (yms_key, timestamp)}
        self.tenantid_cache = OrderedDict()
        self.tenantid_cache_size = cache_size
        self.tenantid_cache_ttl = cache_ttl
        
        # Token缓存：{token: (yms_key, timestamp)}
        self.token_cache = OrderedDict()
        self.token_cache_size = cache_size
        self.token_cache_ttl = cache_ttl

    def _get_from_token_cache(self, token: str) -> Optional[str]:
        """从token缓存中获取 vecdb_code"""
        with self.token_cache_lock:
            if token in self.token_cache:
                vecdb_code, timestamp = self.token_cache[token]
                # 检查是否过期
                if time.time() - timestamp < self.token_cache_ttl:
                    # 更新访问时间（LRU特性）
                    self.token_cache.move_to_end(token)
                    return vecdb_code
                # 如果过期则移除
                del self.token_cache[token]
        return None
    
    def _get_from_tenantid_cache(self, tenantid: str) -> Optional[str]:
        """从tenantid缓存中获取 vecdb_code"""
        with self.tenantid_cache_lock:
            if tenantid in self.tenantid_cache:
                vecdb_code, timestamp = self.tenantid_cache[tenantid]
                # 检查是否过期
                if time.time() - timestamp < self.tenantid_cache_ttl:
                    # 更新访问时间（LRU特性）
                    self.tenantid_cache.move_to_end(tenantid)
                    return vecdb_code
                # 如果过期则移除
                del self.tenantid_cache[tenantid]
        return None

    def _add_to_token_cache(self, token: str, vecdb_code: str):
        """添加token到缓存"""
        with self.token_cache_lock:
            # 添加新条目
            self.token_cache[token] = (vecdb_code, time.time())
            
            # 维护缓存大小
            while len(self.token_cache) > self.token_cache_size:
                self.token_cache.popitem(last=False)

    def _add_to_tenantid_cache(self, tenantid: str, vecdb_code: str):
        """添加token到缓存"""
        with self.tenantid_cache_lock:
            # 添加新条目
            self.tenantid_cache[tenantid] = (vecdb_code, time.time())
            
            # 维护缓存大小
            while len(self.tenantid_cache) > self.tenantid_cache_size:
                self.tenantid_cache.popitem(last=False)

    def get_vecdb_code(self, tenant_id: Optional[str]=None, yht_access_token: Optional[str]=None):
        """
        获取租户路由 vecdb_code 信息
        :param tenant_id: 租户ID
        :param yht_access_token: 认证令牌
        :return: (状态, 路由向量库编码或错误信息)
        """

        if not tenant_id and not yht_access_token:
            logger.error("tenant_id and yht_access_token cannot be empty at the same time")
            return False, "tenant_id and yht_access_token cannot be empty at the same time"

        if yht_access_token:
            vecdb_code = self._get_from_token_cache(yht_access_token)
            if vecdb_code:
                return True, vecdb_code
        elif tenant_id:
            vecdb_code = self._get_from_tenantid_cache(tenant_id)
            if vecdb_code:
                return True, vecdb_code
        
        headers = {"Content-Type": "application/json"}
        params = {
            "type": "logicData"
        }
        if yht_access_token:
            headers['yht_access_token'] = yht_access_token
        else:
            if iuap_request.has_jwt_package:
                headers, params = iuap_request.sign(self.mapping_url, params, headers)
            headers['X-tenantId'] = tenant_id
        last_error = None
        for i in range(1, 3):
            try: 
                resp = requests.request( method="GET", url=self.mapping_url, params=params, verify=False, headers=headers, timeout=self.timeout)
                resp.raise_for_status()
                resp_json = resp.json()
                logger.info("request retry time: %s, url: %s, params: %s, resp: %s", i, self.mapping_url, params, resp_json)
                res_data = resp_json.get("data") if resp_json is not None else None
                if not res_data or not isinstance(res_data, dict):
                    return False, f"get_vecdb_code failed, res_data error, resp:{str(resp_json)}"
                vecdb_code = res_data.get("logicData")
                if vecdb_code:
                    if yht_access_token:
                        self._add_to_token_cache(yht_access_token, vecdb_code)
                    else:
                        self._add_to_tenantid_cache(tenant_id, vecdb_code)
                return True, vecdb_code

            except Exception as e:
                last_error = str(e)
                logger.error("request retry time: %s, url: %s, error: %s", i, self.mapping_url, e)
        return False, f"get_vecdb_code failed, {last_error}"


if __name__ == '__main__':
    print(UserRouter().get_vecdb_code("", ""))


"""
Author: liutengx
Date: 2024-03-05 14:54:05
LastEditTime: 2024-03-05 14:54:05
LastEditors: liutengx
Description: 大模型数据集
FilePath: /iw-algo-fx/intelliw/datasets/datasource_large_model.py
"""
import json
import os
import shutil
import time
from random import Random

from intelliw.config import config
from intelliw.datasets.datasource_base import AbstractDataSource, DataSourceType, DatasetSelector
from intelliw.datasets.spliter import get_set_count, check_split_ratio
from intelliw.utils import iuap_request
from intelliw.utils.exception import DatasetException
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


class CorpusType:
    json = 'json'


class FineTuningSerType:
    train = 'train'
    test = 'test'
    mixed = 'mixed'
    general = 'general'


class DataSourceLMCorpora(AbstractDataSource):
    """
    大模型数据处理
    """
    # dataset_class 一共有 训练 验证 测试 三种数据集
    dataset_class = 3
    # file_name 文件名称
    file_name = ['train.json', 'val.json', 'test.json']
    # 共几个数据集， 当大于1时，写到同一个文件
    dataset_count = 0
    # counts 数据集总行数
    counts = [0] * dataset_class

    def __init__(self, source_address: str, ds_id: str, set_type: str, tenant_id: str = None,
                 download_dir='./tmp_local_large_model_origin_data/'):
        DataSourceLMCorpora.dataset_count += 1
        self.dataset_no = DataSourceLMCorpora.dataset_count

        self.ratio = [config.TRAIN_DATASET_RATIO,
                      config.VALID_DATASET_RATIO, config.TEST_DATASET_RATIO]
        check_split_ratio(*self.ratio)

        self.random = Random(config.DATA_RANDOM_SEED)
        self.download_dir = download_dir

        self.ds_id = ds_id
        self.source_address = source_address
        self.dirpath = [config.LARGE_MODEL_TRAIN_FILEPATH,
                        config.LARGE_MODEL_VAL_FILEPATH, config.LARGE_MODEL_TEST_FILEPATH]
        self.set_type = set_type
        self.tenant_id = tenant_id if tenant_id is not None else config.TENANT_ID

    def total(self):
        return 1

    def reader(self, *args, **kwargs):
        return list()

    def corpora_process(self, split_mode: int = 0):
        """ 大模型数据切分

        Args:
            split_mode (int, optional): 0顺序 1乱序. Defaults to 0.

        此方法中所有长度3的列表, 均代表 [训练, 验证, 测试]
        """
        # 生成数据文件夹
        self._gen_corpora_dir()

        start_time = time.time()
        # 切分函数的缓存, 减少切分次数
        spliter_cache = {}
        # io_reader 数据集写入文件io
        dataset_list = [[], [], []]
        if DataSourceLMCorpora.dataset_count > 1:
            for i in range(DataSourceLMCorpora.dataset_class):
                file_path = os.path.join(self.dirpath[i], self.file_name[i])
                if os.path.exists(file_path):
                    with open(file_path, "r") as fp:
                        dataset_list[i] = json.load(fp)

        # 进行多文件循环数据集拆分
        for epoch, part in enumerate(self._get_reader()):
            # split_mode 根据拆分模式进行处理
            if split_mode:
                self.random.shuffle(part)
            part_count = len(part)

            # set_data 此part数据集原始数据,切分规则按混合数据集[1,0,0],验证数据集[0,1,0]
            set_data = [[]] * DataSourceLMCorpora.dataset_class
            nums = [0] * DataSourceLMCorpora.dataset_class
            if self.set_type == FineTuningSerType.mixed:
                nums = spliter_cache.get(part_count)
                if nums is None:
                    nums = [part_count, 0, 0] if part_count < 10 else get_set_count(
                        part_count, *self.ratio)
                    spliter_cache[part_count] = nums
            elif self.set_type in [FineTuningSerType.train, FineTuningSerType.general]:
                nums = [part_count, 0, 0]
            elif self.set_type == FineTuningSerType.test:
                nums = [0, 0, part_count]

            _start, _end = 0, 0
            for idx, num in enumerate(nums):
                _end += num
                set_data[idx] = part[_start:] if idx == len(nums) - 1 else part[_start:_end]
                _start += num
            part.clear()

            # 更新数据集总数据
            for idx, num in enumerate(nums):
                DataSourceLMCorpora.counts[idx] += num

            [i.extend(s) for i, s in zip(dataset_list, set_data)]

            logger.info(
                f"大模型数据文件下载中: train: {DataSourceLMCorpora.counts[0]} 条, validation: {DataSourceLMCorpora.counts[1]} 条, "
                f"test: {DataSourceLMCorpora.counts[2]} 条")

        # 关闭所有句柄
        for i in range(DataSourceLMCorpora.dataset_class):
            file_path = os.path.join(self.dirpath[i], self.file_name[i])
            with open(file_path, "w", encoding="utf-8") as fp:
                json.dump(dataset_list[i], fp, ensure_ascii=False)

        logger.info(
            f"大模型数据[{self.dataset_no}]下载完成: total: {sum(DataSourceLMCorpora.counts)} 条, "
            f"train: {DataSourceLMCorpora.counts[0]} 条, "
            f"validation: {DataSourceLMCorpora.counts[1]} 条, "
            f"test: {DataSourceLMCorpora.counts[2]} 条, "
            f"time: {time.time() - start_time}")

        # 清空下载文件夹
        shutil.rmtree(self.download_dir, ignore_errors=True)

    def _gen_corpora_dir(self):
        filepath = os.path.join('./', config.LARGE_MODEL_FILEPATH)
        if os.path.exists(filepath) and self.dataset_no == 1:
            logger.warn(f"大模型数据保存路径存在:{filepath}, 正在删除路径内容")
            shutil.rmtree(filepath, ignore_errors=True)
        os.makedirs(config.LARGE_MODEL_TRAIN_FILEPATH, 0o755, True)
        os.makedirs(config.LARGE_MODEL_VAL_FILEPATH, 0o755, True)
        os.makedirs(config.LARGE_MODEL_TEST_FILEPATH, 0o755, True)

    def __call__(self):
        abspath = os.path.abspath('.')
        return [os.path.join(abspath, config.LARGE_MODEL_TRAIN_FILEPATH, self.file_name[0]),
                os.path.join(abspath, config.LARGE_MODEL_VAL_FILEPATH, self.file_name[1]),
                os.path.join(abspath, config.LARGE_MODEL_TEST_FILEPATH, self.file_name[2])]

    # reader
    def _get_reader(self, pagesize=5000):
        return DataSourceLMCorpora.__OnlineReader(self.source_address, self.ds_id, self.tenant_id, pagesize)

    class __OnlineReader:
        def __init__(self, input_address, ds_id, tenant_id, limit=5000):
            """
            eg. 91 elements, page_size = 20, 5 pages as below:
            [0,19][20,39][40,59][60,79][80,90]
            offset 15, limit 30:
            [15,19][20,39][40,44]
            offset 10 limit 5:
            [10,14]
            """
            # self.logger = _get_framework_logger()
            self.input_address = input_address
            self.ds_id = ds_id
            self.page_size = max(100, limit)
            self.page_num = 0
            self.total_read = 0
            self.tenant_id = tenant_id

        @property
        def iterable(self):
            return True

        def __iter__(self):
            return self

        def __next__(self):
            """
            {
                "data": {
                    "content": ["...","..."],
                    "pageNumber": 0,
                    "pageSize": 1,
                    "totalElements": 14,
                    "totalPages": 14
                },
                "msg": "成功",
                "status": 1
            }
            """
            page = None
            try:
                page = self._read_page(self.page_num, self.page_size)
                assert page is not None, "获取的数据为空"
                if page['status'] != 1:
                    raise Exception(page['msg'])
                if page['data']['totalElements'] == 0:
                    raise Exception("大模型 数据获取失败, 无数据")
                content = page['data']['content']
                content = [json.loads(i) for i in content]
            except Exception as e:
                errmsg = f"大模型 数据获取失败, \nds_id: [{self.ds_id}], \nerror: [{e}], \nresponse: {page}"
                raise DatasetException(errmsg)

            if len(content) == 0:
                raise StopIteration
            self.page_num += 1
            self.total_read += len(content)
            return content

        def _read_page(self, page_num, page_size):
            """
            调用智能分析接口，分页读取数据
            :param page_num: 页码，从 0 开始
            :param page_size:  每页大小
            :return:
            """
            request_data = {'dsId': self.ds_id, 'pageNumber': page_num,
                            'pageSize': page_size, 'yTenantId': self.tenant_id}
            response = iuap_request.post_json(url=self.input_address, json=request_data, timeout=config.DATA_SOURCE_READ_TIMEOUT)
            response.raise_for_status()
            return response.json


DatasetSelector.register_func(DataSourceType.LARGE_MODEL, DataSourceLMCorpora, {
    "source_address": ["LARGE_MODEL_PATH", "INPUT_ADDR"],
    "ds_id": "INPUT_DATA_SOURCE_ID",
    "set_type": "FINE_TUNING_SET_TYPE",
    "tenant_id": "TENANT_ID"})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Author: AI Assistant
Date: 2025-08-29
Description: 数据工厂物化表API接口封装
基于物化表接口文档.md的API规范实现
FilePath: /iw-algo-fx/intelliw/datasets/utils/iwfactory_dr_api.py
"""

import json
from typing import Dict, List, Any, Optional, Union
from intelliw.utils import iuap_request
from intelliw.utils.logger import _get_framework_logger
from intelliw.config import config

logger = _get_framework_logger()


class IwFactoryDrAPI:
    """数据工厂物化表API接口封装类"""
    
    def __init__(self, base_url: str = None, tenant_id: str = None, user_id: str = None, user_url_map: Dict[str, str] = None):
        """
        初始化API客户端
        
        Args:
            base_url: API基础URL，如果为None则使用config中的配置
            tenant_id: 租户ID，如果为None则使用config中的配置
            user_id: 用户ID
        """
        self.base_url = base_url or config.DATA_FACTORY_URL

        self.tenant_id = tenant_id or config.TENANT_ID
        self.user_id = user_id
        self.headers = {'service-code': 'iuap-aip-console'}
        
        # API端点映射
        self.endpoints = {
            'std_metadata': '/api/er-std-metadata',  # 表元数据规范
            'physics_table': '/api/er-physics-table',  # 物化建表
            'table_metadata': '/api/er-table-metadata',  # 查看表元数据
            'write_data': '/api/er-write-data',  # 数据批量写入
            'db_tables': '/api/er-db-tables',  # 获取数据源下表集合
            'db_table_columns': '/api/er-db-table-columns',  # 获取数据源下表-列集合
            'select_data': '/api/er-select-data',  # 批量查询数据
        }

        # 用户传入 url 优先使用用户传入的
        self.user_url_map = user_url_map if user_url_map and isinstance(user_url_map, dict) else {}
        if not self.base_url and not self.user_url_map:
            logger.warning('No base_url or user_url_map provided, using default url: %s', config.DATA_FACTORY_URL)

    def _make_request(self, method: str, endpoint: str, data: Dict = None, 
                     headers: Dict = None, params: Dict = None) -> Dict[str, Any]:
        """
        统一的HTTP请求方法
        
        Args:
            method: HTTP方法 ('GET', 'POST')
            endpoint: API端点
            data: 请求数据
            headers: 请求头
            params: URL参数
            
        Returns:
            响应数据字典
        """

        if endpoint.startswith('http://') or endpoint.startswith('https://'):
            url = endpoint 
        elif not self.user_url_map and endpoint in self.user_url_map:
            url = self.user_url_map[endpoint]
        else:
            url = f"{self.base_url}{endpoint}"
        
        # 设置默认请求头
        default_headers = self.headers.copy()
        if self.user_id:
            default_headers['user_id'] = self.user_id

        if headers:
            default_headers.update(headers)
        
        try:
            if method.upper() == 'GET':
                response = iuap_request.get(
                    url, params=params, headers=default_headers,
                    timeout=config.DATA_SOURCE_READ_TIMEOUT,
                    auth_type=iuap_request.AuthType.YHT
                )
            elif method.upper() == 'POST':
                response = iuap_request.post_json(
                    url, json=data, headers=default_headers,
                    timeout=config.DATA_SOURCE_READ_TIMEOUT,
                    auth_type=iuap_request.AuthType.YHT
                )
            else:
                raise ValueError(f"不支持的HTTP方法: {method}")
            response.raise_for_status()
            if response.status != 200:
                raise Exception(f"API请求失败，状态码: {response.status}, URL: {url}")
            
            result = response.json
            # 历史返回状态为0时为正常返回,最新返回200状态码，则返回正常结果，同时兼容0和200状态码
            if result.get('code') != 0 and result.get('code') != 200:
                raise Exception(f"API返回错误，code: {result.get('code')}, result: {result}")
            
            return result
            
        except Exception as e:
            logger.error(f"API请求异常: {str(e)}")
            raise
    
    def std_metadata(self, datasource_id: str, columns: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        表元数据规范API
        
        Args:
            datasource_id: 数据源ID
            columns: 字段列表，每个字段包含：
                - columnEnglishName: 字段英文名称
                - columnDescribe: 字段描述
                - columnType: 字段类型
                - columnLength: 字段长度
                - numericScale: 小数长度
                - columnNullFlag: 是否为空（1是空 0不是空）
                - columnKeyFlag: 是否主键（1是主键 0不是主键）
                
        Returns:
            规范后的字段元数据
        """
        data = {
            "ytenantId": self.tenant_id,
            "datasourceId": datasource_id,
            "columns": columns
        }

        if 'std_metadata' in self.user_url_map:
            endpoint = self.user_url_map['std_metadata']
        else:
            endpoint = self.endpoints['std_metadata']
        
        return self._make_request('POST', endpoint, data=data)
    
    def create_physics_table(self, datasource_id: str, project_name: str, 
                           table_name: str, description: str, 
                           columns: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        物化建表API
        
        Args:
            datasource_id: 数据源ID
            project_name: 项目英文名称
            table_name: 表名
            description: 表描述
            columns: 字段列表，格式同std_metadata
            
        Returns:
            包含logicTableId的响应数据
        """
        data = {
            "ytenantId": self.tenant_id,
            "projectName": project_name,
            "datasourceId": datasource_id,
            "table": {
                "tableName": table_name,
                "description": description
            },
            "columns": columns
        }

        if 'physics_table' in self.user_url_map:
            endpoint = self.user_url_map['physics_table']
        else:
            endpoint = self.endpoints['physics_table']
        
        return self._make_request('POST', endpoint, data=data)
    
    def get_table_metadata(self, logic_table_id: str) -> Dict[str, Any]:
        """
        查看表元数据API
        
        Args:
            logic_table_id: 逻辑模型ID
            
        Returns:
            表元数据信息
        """
        if 'table_metadata' in self.user_url_map:
            endpoint = self.user_url_map['table_metadata']
        else:
            endpoint = f"{self.endpoints['table_metadata']}/{self.tenant_id}/{logic_table_id}"

        return self._make_request('GET', endpoint)
    
    def write_data(self, datasource_id: str, table_name: str, 
                   column_schema: List[str], model_data: List[List[Any]],
                   logic_table_id: str = None) -> Dict[str, Any]:
        """
        数据批量写入API（支持任意数据源表）
        
        Args:
            datasource_id: 数据源ID
            table_name: 表名
            column_schema: 字段名列表
            model_data: 数据值列表（二维数组）
            logic_table_id: 逻辑模型ID（可选，优先使用ER表）
            
        Returns:
            写入结果
        """
        data = {
            "ytenantId": self.tenant_id,
            "datasourceId": datasource_id,
            "dataModel": {
                "tableName": table_name,
                "columnSchema": column_schema,
                "modelData": model_data
            }
        }
        
        if logic_table_id:
            data["logicTableId"] = logic_table_id
        
        if 'write_data' in self.user_url_map:
            endpoint = self.user_url_map['write_data']
        else:
            endpoint = self.endpoints['write_data']
        
        return self._make_request('POST', endpoint, data=data)
    
    def get_db_tables(self, datasource_id: str) -> Dict[str, Any]:
        """
        获取数据源下表集合API
        
        Args:
            datasource_id: 数据源ID
            
        Returns:
            表列表
        """
        if 'db_tables' in self.user_url_map:
            endpoint = self.user_url_map['db_tables']
        else:
            endpoint = f"{self.endpoints['db_tables']}/{self.tenant_id}/{datasource_id}"

        return self._make_request('GET', endpoint)
    
    def get_db_table_columns(self, table_id: str) -> Dict[str, Any]:
        """
        获取数据源下表-列集合API
        
        Args:
            table_id: 表ID
            
        Returns:
            列列表
        """

        if 'db_table_columns' in self.user_url_map:
            endpoint = self.user_url_map['db_table_columns']
        else:
            endpoint = f"{self.endpoints['db_table_columns']}/{self.tenant_id}/{table_id}"
        return self._make_request('GET', endpoint)
    
    def select_data(self, datasource_id: str, table_name: str,
                   column_schema: List[str] = None, 
                   page: Dict[str, int] = None,
                   where_condition: List[Dict[str, Any]] = None,
                   where_relation: List[str] = None) -> Dict[str, Any]:
        """
        批量查询数据API
        
        Args:
            datasource_id: 数据源ID
            table_name: 表名
            column_schema: 展示字段列表（可选，空数组表示获取所有字段）
            page: 分页条件，包含：
                - currentPage: 当前页，从1开始
                - pageSize: 页大小
                - totalCount: 总记录数（可选）
                - totalPage: 总页数（可选）
            where_condition: where条件列表，每个条件包含：
                - columnSchema: 字段名
                - columnValues: 值列表
                - operator: 操作符（=, <>, !=, <, <=, >, >=, like, not like, is null, is not null, in）
            where_relation: 关联关系列表（可选，支持and, or）
            
        Returns:
            查询结果，包含columnSchema和modelData
        """
        data = {
            "ytenantId": self.tenant_id,
            "logicTableId": "",  # 根据文档，空字符串表示任意数据源表
            "datasourceId": datasource_id,
            "dataModel": {
                "tableName": table_name,
                "columnSchema": column_schema or [],
                "page": page or {
                    "currentPage": 1,
                    "pageSize": 1000,
                    "totalCount": 0,
                    "totalPage": 0
                },
                "where": {
                    "condition": where_condition or [],
                    "relation": where_relation or []
                }
            }
        }

        if 'select_data' in self.user_url_map:
            endpoint = self.user_url_map['select_data']
        else:
            endpoint = self.endpoints['select_data']
        
        return self._make_request('POST', endpoint, data=data)
    
    def select_data_paginated(self, datasource_id: str, table_name: str,
                            column_schema: List[str] = None,
                            page_size: int = 1000,
                            current_page: int = 1,
                            where_condition: List[Dict[str, Any]] = None,
                            where_relation: List[str] = None) -> Dict[str, Any]:
        """
        分页查询数据的便捷方法
        
        Args:
            datasource_id: 数据源ID
            table_name: 表名
            column_schema: 展示字段列表
            page_size: 每页大小
            current_page: 当前页码
            where_condition: where条件
            where_relation: 关联关系
            
        Returns:
            分页查询结果
        """
        page = {
            "currentPage": current_page,
            "pageSize": page_size,
            "totalCount": 0,
            "totalPage": 0
        }
        
        return self.select_data(
            datasource_id=datasource_id,
            table_name=table_name,
            column_schema=column_schema,
            page=page,
            where_condition=where_condition,
            where_relation=where_relation
        )
    
    def select_all_data(self, datasource_id: str, table_name: str,
                       column_schema: List[str] = None,
                       batch_size: int = 1000,
                       where_condition: List[Dict[str, Any]] = None,
                       where_relation: List[str] = None) -> List[List[Any]]:
        """
        获取所有数据的便捷方法（自动分页获取）
        
        Args:
            datasource_id: 数据源ID
            table_name: 表名
            column_schema: 展示字段列表
            batch_size: 每批获取的数据量
            where_condition: where条件
            where_relation: 关联关系
            
        Returns:
            所有数据的二维数组
        """
        all_data = []
        current_page = 1
        
        while True:
            result = self.select_data_paginated(
                datasource_id=datasource_id,
                table_name=table_name,
                column_schema=column_schema,
                page_size=batch_size,
                current_page=current_page,
                where_condition=where_condition,
                where_relation=where_relation
            )
            
            data_model = result.get('data', {}).get('dataModel', {})
            model_data = data_model.get('modelData', [])
            
            if not model_data:
                break
                
            all_data.extend(model_data)
            
            # 检查是否还有更多数据
            page_info = data_model.get('page', {})
            total_page = page_info.get('totalPage', 1)
            if current_page >= total_page:
                break
                
            current_page += 1
        
        return all_data


# 全局API实例
_iwfactory_api = None


def get_iwfactory_api() -> IwFactoryDrAPI:
    """获取全局IwFactoryDrAPI实例"""
    global _iwfactory_api
    if _iwfactory_api is None:
        _iwfactory_api = IwFactoryDrAPI()
    return _iwfactory_api


def set_iwfactory_api_config(base_url: str = None, tenant_id: str = None, user_id: str = None):
    """设置全局API配置"""
    global _iwfactory_api
    _iwfactory_api = IwFactoryDrAPI(base_url, tenant_id, user_id)


# 便捷函数封装
def std_metadata(datasource_id: str, columns: List[Dict[str, Any]]) -> Dict[str, Any]:
    """表元数据规范API的便捷函数"""
    return get_iwfactory_api().std_metadata(datasource_id, columns)


def create_physics_table(datasource_id: str, project_name: str, 
                        table_name: str, description: str, 
                        columns: List[Dict[str, Any]]) -> Dict[str, Any]:
    """物化建表API的便捷函数"""
    return get_iwfactory_api().create_physics_table(
        datasource_id, project_name, table_name, description, columns)


def get_table_metadata(logic_table_id: str) -> Dict[str, Any]:
    """查看表元数据API的便捷函数"""
    return get_iwfactory_api().get_table_metadata(logic_table_id)


def write_data(datasource_id: str, table_name: str, 
               column_schema: List[str], model_data: List[List[Any]],
               logic_table_id: str = None) -> Dict[str, Any]:
    """数据批量写入API的便捷函数"""
    return get_iwfactory_api().write_data(
        datasource_id, table_name, column_schema, model_data, logic_table_id)


def get_db_tables(datasource_id: str) -> Dict[str, Any]:
    """获取数据源下表集合API的便捷函数"""
    return get_iwfactory_api().get_db_tables(datasource_id)


def get_db_table_columns(table_id: str) -> Dict[str, Any]:
    """获取数据源下表-列集合API的便捷函数"""
    return get_iwfactory_api().get_db_table_columns(table_id)


def select_data(datasource_id: str, table_name: str,
               column_schema: List[str] = None, 
               page: Dict[str, int] = None,
               where_condition: List[Dict[str, Any]] = None,
               where_relation: List[str] = None) -> Dict[str, Any]:
    """批量查询数据API的便捷函数"""
    return get_iwfactory_api().select_data(
        datasource_id, table_name, column_schema, page, where_condition, where_relation)


def select_all_data(datasource_id: str, table_name: str,
                   column_schema: List[str] = None,
                   batch_size: int = 1000,
                   where_condition: List[Dict[str, Any]] = None,
                   where_relation: List[str] = None) -> List[List[Any]]:
    """获取所有数据的便捷函数"""
    return get_iwfactory_api().select_all_data(
        datasource_id, table_name, column_schema, batch_size, where_condition, where_relation)


if __name__ == "__main__":
    # 测试代码
    import os
    
    # 设置测试配置
    os.environ.setdefault('DATA_FACTORY_URL', 'http://localhost:8080')
    os.environ.setdefault('TENANT_ID', 'test_tenant')
    
    # 创建API实例
    api = IwFactoryDrAPI()
    
    # 测试表元数据规范
    test_columns = [
        {
            "columnEnglishName": "id",
            "columnDescribe": "用户编号",
            "columnType": "INT",
            "columnLength": 11,
            "numericScale": 0,
            "columnNullFlag": 1,
            "columnKeyFlag": 1
        },
        {
            "columnEnglishName": "name",
            "columnDescribe": "用户名称",
            "columnType": "VARCHAR",
            "columnLength": 50,
            "numericScale": 0,
            "columnNullFlag": 1,
            "columnKeyFlag": 0
        }
    ]
    
    try:
        result = api.std_metadata("test_datasource", test_columns)
        print("表元数据规范测试成功:", result)
    except Exception as e:
        print("表元数据规范测试失败:", str(e))
    
    print("IwFactoryDrAPI 初始化完成")